/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */


/*!
    \page lcmaps_c_pep.mod C PEP for Argus plugin

    \section posixsynopsis SYNOPSIS
    \b lcmaps_c_pep.mod

    \section posixdesc DESCRIPTION

The GUMS module will connect to the GUMS server using the SAML protocol (provided by the Privilege Project) and return a Unix UID, GID and possibly secondary GIDs to the framework.


*/


/*!
    \file   lcmaps_c_pep.c
    \brief  Plugin to connect to a PEPd and return a UID, a GID and secondary GIDs.
    \author Oscar Koeroo for the EGEE.
*/


/* In any case, also define _XOPEN_SOURCE */
#define _XOPEN_SOURCE	600



/*****************************************************************************
                            Include header files
******************************************************************************/
#include "lcmaps_c_pep_config.h"

/* Note: can only now define the _GNU_SOURCE since we have included the config
 * header file */
#ifdef HAVE_GETRESUID
# define _GNU_SOURCE
#endif

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>
#include <pwd.h>
#include <grp.h>
#include <time.h>

#include <netdb.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>
#include <fcntl.h>

#include <lcmaps/lcmaps_modules.h>
#include <lcmaps/lcmaps_arguments.h>
#include <lcmaps/lcmaps_cred_data.h>
#include <lcmaps/lcmaps_vo_data.h>

#include <lcmaps/lcmaps_account.h>

#if defined(HAVE_LCMAPS_LCMAPS_PLUGIN_PROTOTYPES_H)
#   include <lcmaps/lcmaps_plugin_prototypes.h>
#else
#   include "lcmaps_plugin_prototypes.h"
#endif

#include "pep-c-log.h"
#include "pep-c-obligation-handlers.h"
#include "pep-c-obligation-handlers_helpers.h"

#include <openssl/bio.h>
#include <openssl/x509.h>
#include <openssl/pem.h>



/******************************************************************************
                                Definitions
******************************************************************************/

#define PLUGIN_RUN      0
#define PLUGIN_VERIFY   1


#define LCMAPS_MAX_URNURL_LENGTH 512

#define MAX_ENDPOINTS         32
#define MAX_ENDPOINT_BUFSIZE  ((size_t)256)
#define MAX_PWLEN             2048

#define PEP_OPTION_SSL_VALIDATION_DEFAULT -1
#define PEP_OPTION_SSL_VALIDATION_OFF 0
#define PEP_OPTION_SSL_VALIDATION_ON 1


#ifndef PEP_OPTION_SSL_CIPHER_LIST_DEFAULT
    #define PEP_OPTION_SSL_CIPHER_LIST_DEFAULT NULL
#endif /* PEP_OPTION_SSL_CIPHER_LIST_DEFAULT */

#define PEP_OPTION_CLIENT_CERT_UNDEFINED 0
#define PEP_OPTION_CLIENT_CERT_IMPLICIT  1
#define PEP_OPTION_CLIENT_CERT_EXPLICIT  2
#define PEP_OPTION_CLIENT_CERT_NONE      3

typedef enum ENDPOINT_SELECTION_E {
    ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM = 1,
    ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START  = 2,
    ENDPOINT_SELECTION_RANDOM                    = 3
} ENDPOINT_SELECTION_T;

#define PROFILE_BOTH	"both"
static int    use_profile = PROFILE_SWITCH_GRIDWN; /* Use only the new Grid WN
						      profile from Switch as
						      default */

static ENDPOINT_SELECTION_T endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START;
static char   pep_daemon_endpoint_urls[MAX_ENDPOINTS][MAX_ENDPOINT_BUFSIZE];
static unsigned int n_endpoints = 0;
static long   endpoint_retrycount = 2; /* retry each endpoint twice */
static int    endpoint_retrycount_is_set = 0;

static int    pep_option_ssl_validation = PEP_OPTION_SSL_VALIDATION_DEFAULT;
static char * pep_option_ssl_cipher_list = PEP_OPTION_SSL_CIPHER_LIST_DEFAULT; /* Default must be overriden by the configuration to work on a Red Hat distro with libcurl/libnss */
static char * pep_option_server_cert = NULL;
static char * pep_option_server_capath = NULL;
static char * pep_option_server_capath_default = NULL;
static char * pep_option_client_cert = NULL;
static char * pep_option_client_key = NULL;
static char * pep_option_client_keypassword = NULL;
static char * pep_option_client_keypasswordfile = NULL;
static int    pep_option_client_cert_mode = PEP_OPTION_CLIENT_CERT_UNDEFINED;
static int    pep_option_use_pilot_proxy_as_cafile = 0;
static int    pep_option_send_pilot_cert_chain = 0;
static char   keypasswordbuffer[MAX_PWLEN];


static const char * resourcetype      = NULL;
static const char * actiontype        = NULL;
static char * resourceid              = NULL;
static char * actionid                = NULL;

/* Loglevel for the pep-api-c library logging */
static int    pep_c_loglevel             = PEP_LOGLEVEL_WARN;

/******************************************************************************
                          Module specific prototypes
******************************************************************************/

static int plugin_run_or_verify(int, lcmaps_argument_t *, int);

static char *plugin_c_pep_readpwfile(char *fn, char *buf, int n);
static int plugin_c_pep_optequal(const char *s, const char *t);
static char *plugin_c_pep_set_capath_default(char **p);
static char *lcmaps_plugin_c_pep_cleanse_pemstring(char *in, char **destp);

static int downgradeEffectiveToRealUid (uid_t * real_uid, uid_t * saved_uid);
static int upgradeEffectiveToRealUid (uid_t real_uid, uid_t saved_uid);

static char * readProxyFile (const char * proxyfile);
static char * lcmaps_stackofx509_to_pem (STACK_OF(X509) * px509_chain);


/******************************************************************************
Function:   plugin_initialize
Description:
    Initialize plugin
Parameters:
    argc, argv
    argv[0]: the name of the plugin
Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
    LCMAPS_MOD_NOFILE  : db file not found (will halt LCMAPS initialization)
******************************************************************************/
int plugin_initialize(int argc, char **argv)
{
    const char * logstr = "lcmaps_plugin_c_pep-plugin_initialize()";
    int    i = 0;

    lcmaps_log(LOG_DEBUG,"%s: passed arguments:\n", logstr);
    for (i=0; i < argc; i++)
    {
        lcmaps_log(LOG_DEBUG,"%s: arg %d is %s\n", logstr, i, argv[i]);
    }

    /*
     * Parse arguments, argv[0] = name of plugin, so start with i = 1
     */
    for (i = 1; i < argc; i++)
    {
        /* Endpoint selection strategy */
	if ( plugin_c_pep_optequal(argv[i],"endpoint-strategy")
	     && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if      (strcasecmp (argv[i + 1], "round-robin") == 0)              endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM;
                else if (strcasecmp (argv[i + 1], "round-robin-random-start") == 0) endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START;
                else if (strcasecmp (argv[i + 1], "random") == 0)                   endpoint_selection_strategy = ENDPOINT_SELECTION_RANDOM;
                else
                {
                    lcmaps_log (LOG_ERR, "%s: Error: --endpoint-strategy not understood. Options are: \"round-robin\", \"round-robin-random-start\" and \"random\".\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        /* endpoints */
        else if ( plugin_c_pep_optequal(argv[i], "pep-daemon-endpoint-url")
             && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (n_endpoints < MAX_ENDPOINTS)
                {
/*                    bzero (pep_daemon_endpoint_urls[n_endpoints], MAX_ENDPOINT_BUFSIZE);*/
		    memset(pep_daemon_endpoint_urls[n_endpoints], '\0', MAX_ENDPOINT_BUFSIZE);
                    if (strlen (argv[1 + 1]) >= MAX_ENDPOINT_BUFSIZE)
                    {
                        lcmaps_log (LOG_ERR, "%s: Error: endpoint URL is too long: %s...\n", strncpy(pep_daemon_endpoint_urls[n_endpoints], argv[i + 1], MAX_ENDPOINT_BUFSIZE - 1), logstr);
                        return LCMAPS_MOD_FAIL;
                    }
                    else
                    {
                        strncpy(pep_daemon_endpoint_urls[n_endpoints], argv[i + 1], MAX_ENDPOINT_BUFSIZE - 1);
                        n_endpoints++;
                    }
                }
                else
                {
                    lcmaps_log (LOG_ERR, "%s: Error: Too many endpoints configured, the maximum amount of configurable endpoints is %d\n", logstr, MAX_ENDPOINTS);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

	/* Endpoint retry count */
        else if ( plugin_c_pep_optequal(argv[i], "retry")
		&& (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
		errno=0;
                endpoint_retrycount = strtol (argv[i + 1], NULL, 10);
                if (errno == EINVAL)
                {
                    lcmaps_log(LOG_ERR, "%s: Error: Couldn't parse the value \"%s\" to a number for the --retry option.\n", logstr, argv[i + 1]);
                    return LCMAPS_MOD_FAIL;
                }
                if (endpoint_retrycount<=0)
                {
                    lcmaps_log(LOG_ERR, "%s: Error: Endpoint retry count to low. It must be at least '1'\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
		endpoint_retrycount_is_set=1;
            }
            i++;
        }

        /* Resource Type - Set the resource type to CE, WN, RB, SE */
        else if ( plugin_c_pep_optequal(argv[i], "resourcetype")
                  && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                char * tmprt = argv[i + 1];

                if (strcasecmp (tmprt, "se") == 0)      resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_SE;
                else if (strcasecmp (tmprt, "ce") == 0) resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_CE;
                else if (strcasecmp (tmprt, "wn") == 0) resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_WN;
                else if (strcasecmp (tmprt, "rb") == 0) resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_RB;
                else
                {
                    lcmaps_log (LOG_ERR, "%s: Error: Resource type unknown: %s\n", logstr, tmprt);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        /* Action Type - Set the action type to CE, WN, RB, SE */
        else if ( plugin_c_pep_optequal(argv[i], "actiontype")
                  && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                char * tmprt = argv[i + 1];

                if (strcasecmp (tmprt, "access-file") == 0)      actiontype = XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_ACCESS_FILE;
                else if (strcasecmp (tmprt, "queue") == 0)       actiontype = XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_QUEUE;
                else if (strcasecmp (tmprt, "execute-now") == 0) actiontype = XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_EXEC_NOW;
                else
                {
                    lcmaps_log (LOG_ERR, "%s: Error: Action type unknown: %s\n", logstr, tmprt);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        /* Resource ID - Override of the resource id. XACML spec dictates this to be changeable */
        else if ( plugin_c_pep_optequal(argv[i], "resourceid")
                  && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > LCMAPS_MAX_URNURL_LENGTH)
                {
                    lcmaps_log (LOG_ERR, "%s: Error: Resource ID URL or URN to long.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    resourceid = argv[i + 1];
            }
            i++;
        }

        /* Action ID - Override of the action id. XACML spec dictates this to be changeable */
        else if ( plugin_c_pep_optequal(argv[i], "actionid") && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > LCMAPS_MAX_URNURL_LENGTH)
                {
                    lcmaps_log (LOG_ERR, "%s: Error: Resource ID URL or URN to long.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    actionid = argv[i + 1];
            }
            i++;
        }

	/* Next one is deprecated: use pep-c-loglevel instead. If someone uses
	 * it, 'translate' to setting --pep-c-loglevel to debug */
        /* Debug setting for the PEP-C library (will log on stderr) */
        else if ( plugin_c_pep_optequal(argv[i], "pep-c-debug") )
        {
	    lcmaps_log (LOG_WARNING, "%s: Warning: using deprecated debug flag: use --pep-c-loglevel instead\n", __func__);
	    pep_c_loglevel = PEP_LOGLEVEL_DEBUG;
        }
	else if ( plugin_c_pep_optequal(argv[i], "pep-c-loglevel") 
		  && (i + 1 < argc) )
	{
	    if (argv[i+1]==NULL)    {
		lcmaps_log(LOG_ERR,
			"%s: Missing value for \"--pep-c-loglevel\" option\n",
			__func__);
		return LCMAPS_MOD_FAIL;
	    }
	    if (strcmp(argv[i+1],"error")==0)
		pep_c_loglevel=PEP_LOGLEVEL_ERROR;
	    else if (strcmp(argv[i+1],"warning")==0)
		pep_c_loglevel=PEP_LOGLEVEL_WARN;
	    else if (strcmp(argv[i+1],"info")==0)
		pep_c_loglevel=PEP_LOGLEVEL_INFO;
	    else if (strcmp(argv[i+1],"debug")==0)
		pep_c_loglevel=PEP_LOGLEVEL_DEBUG;
	    else {
		lcmaps_log(LOG_ERR,
			"%s: Invalid value for \"--pep-c-loglevel\" option: \"%s\"\n",
			__func__, argv[i+1]);
		return LCMAPS_MOD_FAIL;
	    }
	    i++;
	}

	/* When set, we will not check whether there are unhandled obligations
	 * after all obligation handlers have run. This goes against the specs,
	 * but might be necessary in case of non-conforming services */
	else if (plugin_c_pep_optequal(argv[i], "override-unhandled-obligations-check") )
	{
	    /* Call setter function from pep-c-obligation-handlers_helpers.c */
	    lcmaps_log (LOG_WARNING,
		    "%s: Warning: overriding the checking of unhandled obligations.\n", logstr);
	    set_ignore_unhandled_obligations(1);
	}

	/* Use a different SSL cipherlist. Useful to overcome the difference
	 * between OpenSSL/GnuTLS, libnss and others. */
        else if ( plugin_c_pep_optequal(argv[i], "ssl-cipher-list")
                  && (i + 1 < argc) )
        {
            if (argv[i + 1] != NULL)
            {
                pep_option_ssl_cipher_list = argv[i + 1];
            }
            i++;
        }

        /* force the PEP-C library to do SSL validation checks, regardless of its default setting */
        else if ( plugin_c_pep_optequal(argv[i],  "check-certificates") )
        {
            pep_option_ssl_validation = PEP_OPTION_SSL_VALIDATION_ON;
        }

        /* force the PEP-C library to ignore SSL validation checks */
        else if ( plugin_c_pep_optequal(argv[i], "no-check-certificates") )
        {
            pep_option_ssl_validation = PEP_OPTION_SSL_VALIDATION_OFF;
        }

        /* CApath */
        else if ( ( plugin_c_pep_optequal(argv[i], "capath") || plugin_c_pep_optequal(argv[i], "CApath") )
                  && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > PATH_MAX)
                {
                    lcmaps_log (LOG_ERR, "%s: Error: CApath too long in argument.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    pep_option_server_capath = argv[i + 1];
            }
            i++;
        }
        /* file with a CA certificate or set of certs */
        else if ( plugin_c_pep_optequal(argv[i], "cafile") && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > PATH_MAX)
                {
                    lcmaps_log (LOG_ERR, "%s: Error: ca file name too long in argument.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    pep_option_server_cert = argv[i + 1];
            }
            i++;
        }

        /* file with a CA certificate or set of certs */
        else if ( plugin_c_pep_optequal(argv[i], "use-pilot-proxy-as-cafile") )
        {
            pep_option_use_pilot_proxy_as_cafile = 1;
        }

        /* treat a Not Applicable as a success, this is necessary when Argus is
	 * only used for a banning list */
        else if ( plugin_c_pep_optequal(argv[i], "treat-notapplicable-as-success") ||
		  plugin_c_pep_optequal(argv[i], "banning-only-mode") )
        {
	    /* Call setter function from pep-c-obligation-handlers_helpers.c */
	    lcmaps_log (LOG_DEBUG, "%s: Running in 'banning-only-mode', "
				   "treating a 'Not Applicable' as success.\n",
				   logstr);
	    set_treat_notapplicable_as_success(1);
        }

        /* use implicit user proxy, explicit file or no client cert at all */
        else if ( plugin_c_pep_optequal(argv[i], "pep-certificate-mode") && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                char * tmprt = argv[i + 1];

                if (strcasecmp (tmprt, "implicit") == 0)      pep_option_client_cert_mode = PEP_OPTION_CLIENT_CERT_IMPLICIT;
                else if (strcasecmp (tmprt, "explicit") == 0) pep_option_client_cert_mode = PEP_OPTION_CLIENT_CERT_EXPLICIT;
                else if (strcasecmp (tmprt, "none") == 0)     pep_option_client_cert_mode = PEP_OPTION_CLIENT_CERT_NONE;
                else
                {
                    lcmaps_log (LOG_ERR, "%s: Error: pep certificate mode unknown: %s (should be implicit, explicit, or none)\n", logstr, tmprt);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }
	/* certificate to use to contact PEPd (other curl SSL settings) */
        else if ( plugin_c_pep_optequal(argv[i], "certificate") && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > PATH_MAX)
                {
                    lcmaps_log (LOG_ERR, "%s: Error: certificate file name too long in argument.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    pep_option_client_cert = argv[i + 1];
            }
            i++;
        }

	/* certificate to use to contact PEPd (other curl SSL settings) */
        else if ( plugin_c_pep_optequal(argv[i], "key") && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > PATH_MAX)
                {
                    lcmaps_log (LOG_ERR, "%s: Error: certificate key name too long in argument.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    pep_option_client_key = argv[i + 1];
            }
            i++;
        }

	/* certificate to use to contact PEPd (other curl SSL settings) */
        else if ( plugin_c_pep_optequal(argv[i], "passfile") && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > PATH_MAX)
                {
                    lcmaps_log (LOG_ERR, "%s: Error: passwork file name too long in argument.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    pep_option_client_keypasswordfile = argv[i + 1];
            }
            i++;
        }

	/* certificate to use to contact PEPd (other curl SSL settings) */
        else if ( plugin_c_pep_optequal(argv[i], "pass") && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > (size_t)MAX_PWLEN)
                {
                    lcmaps_log (LOG_ERR, "%s: Error: password too long in argument.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    pep_option_client_keypassword = argv[i + 1];
            }
            i++;
        }

	/* Profile choice */
        else if ( plugin_c_pep_optequal(argv[i], "profile") && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strcasecmp(argv[i + 1], XACML_GRIDWN_PROFILE_VERSION) == 0)
                {
                    use_profile = PROFILE_SWITCH_GRIDWN;
                }
                else if (strcasecmp(argv[i + 1], XACML_INTEROP_PROFILE_VERSION) == 0)
                {
                    use_profile = PROFILE_AUTHZ_INTEROP;
                }
		else if (strcasecmp(argv[i + 1], PROFILE_BOTH) == 0)
		{
		    use_profile = PROFILE_SWITCH_GRIDWN | PROFILE_AUTHZ_INTEROP;
		}
                else
                {
                    lcmaps_log (LOG_ERR, "%s: Error: the configured profile is not supported. Supported profiles are: \"%s\"(default) and \"%s\", or special value \"%s\".\n", logstr, XACML_GRIDWN_PROFILE_VERSION, XACML_INTEROP_PROFILE_VERSION, PROFILE_BOTH);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

	/* Whether to send the pilot cert-chain in the environment */
        else if ( plugin_c_pep_optequal(argv[i], "send-pilot-cert-chain") )
        {
            pep_option_send_pilot_cert_chain = 1;
        }

        /* Parameter unknown */
        else
        {
            lcmaps_log (LOG_ERR,"%s: Error in initialization parameter: %s (failure)\n", logstr, argv[i]);
            return LCMAPS_MOD_FAIL;
        }
    }

    /*************************************************************
     * Do sanity checks on the provided input
     *************************************************************/

    /* Do we have at least one endpoint? */
    if (n_endpoints == 0)	{
        lcmaps_log (LOG_ERR, "%s: Error: need at least one endpoint.\n", logstr);
        return LCMAPS_MOD_FAIL;
    }

    /* Default *overall* retrycount, if not set on commandline:
     * - For multiple endpoints set to 1, i.e. each endpoint is tried once,
     * - For one endpoint, set to two, i.e. this one endpoint is tried twice. */
    if (!endpoint_retrycount_is_set)
	endpoint_retrycount = (n_endpoints > 1 ? 1 : 2);

    /* If no cert-mode was specified, but an explicit cert was given, assume
     * this was intended */
    if ( ( PEP_OPTION_CLIENT_CERT_UNDEFINED == pep_option_client_cert_mode )
         && ( pep_option_client_cert || pep_option_client_key ) ) {
	pep_option_client_cert_mode = PEP_OPTION_CLIENT_CERT_EXPLICIT;
    }
    /* Otherwise, the default is 'none' to make it work without config for
     * glexec-on-WNs, but in that caase you will not have client auth (anyone
     * can impersone anybody). If you want client auath in the glexec-on-WN
     * scenario, you need the implicit mode, but this may break when the proxy
     * is on a NFS root-squashed file system. But fixing that bug is disallowed
     * by policy. */
    if ( PEP_OPTION_CLIENT_CERT_UNDEFINED == pep_option_client_cert_mode )  {
	pep_option_client_cert_mode = PEP_OPTION_CLIENT_CERT_NONE;
    }

    /* Basic sanity checks on the certificate file specification that only makes
     * sense in explicit mode */
    if ( pep_option_client_cert && ( PEP_OPTION_CLIENT_CERT_EXPLICIT != pep_option_client_cert_mode ) ) {
        lcmaps_log (LOG_ERR, "%s: Error: certificate file must be used in explicit mode only.\n", logstr);
        return LCMAPS_MOD_FAIL;
    }
    if ( ( PEP_OPTION_CLIENT_CERT_EXPLICIT == pep_option_client_cert_mode ) &&
         ( (!pep_option_client_cert) || (!pep_option_client_key) ) ) {
        lcmaps_log (LOG_ERR, "%s: Error: explicit certificate mode requires a certificate file and a key.\n", logstr);
        return LCMAPS_MOD_FAIL;
    }
#if 0
    /* GGUS ticket: 72752
     * Disabling the requirement to have an X509_USER_PROXY in the plugin
     * initialization phase*/
    if ( ( PEP_OPTION_CLIENT_CERT_IMPLICIT == pep_option_client_cert_mode ) && !getenv("X509_USER_PROXY") ) {
        lcmaps_log (LOG_ERR, "%s: Error: PEPd communication in implicit mode requires proxy filename in X509_USER_PROXY.\n", logstr);
        return LCMAPS_MOD_FAIL;
    }
#endif
    
    /* --use-pilot-proxy-as-cafile and --cafile are mutually exclusive */
    if ( pep_option_use_pilot_proxy_as_cafile && pep_option_server_cert!=NULL )   {
	lcmaps_log (LOG_ERR,
		"%s: Error: --cafile and --use-pilot-proxy-as-cafile are mutually exclusive.\n", logstr);
	return LCMAPS_MOD_FAIL;
    }
    
    /* --use-pilot-proxy-as-cafile needs PEP_OPTION_CLIENT_CERT_IMPLICIT */
    if ( pep_option_use_pilot_proxy_as_cafile &&
	 PEP_OPTION_CLIENT_CERT_IMPLICIT != pep_option_client_cert_mode )   {
	lcmaps_log (LOG_ERR,
		"%s: Error: --use-pilot-proxy-as-cafile needs implicit certificate mode.\n", logstr);
	return LCMAPS_MOD_FAIL;
    }

    lcmaps_log(LOG_DEBUG, "%s: PEPd client certificate mode set to %d.\n", logstr, pep_option_client_cert_mode);


    /* password should either be in a file OR in the command line, not both */
    if ( pep_option_client_keypassword && pep_option_client_keypasswordfile ) {
        lcmaps_log (LOG_ERR, "%s: Error: PEPd client key password should either be in a file OR in the command line, not both.\n", logstr);
        return LCMAPS_MOD_FAIL;
    }
    if ( pep_option_client_keypasswordfile ) {
      pep_option_client_keypassword = plugin_c_pep_readpwfile(pep_option_client_keypasswordfile,keypasswordbuffer,MAX_PWLEN);
      if ( !pep_option_client_keypassword ) {
        lcmaps_log (LOG_ERR, "%s: Error: PEPd client key password could not be read from file %s.\n", logstr, pep_option_client_keypasswordfile);
        return LCMAPS_MOD_FAIL;
      }
    }

    /* the CApath is taken from the command line argument in
     * pep_option_server_capath if set, but otherwise defaults
     */
    if ( !pep_option_server_capath ) {
      pep_option_server_capath = plugin_c_pep_set_capath_default(&pep_option_server_capath_default);
    }


    switch (endpoint_selection_strategy)
    {
        case ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM :
            lcmaps_log(LOG_DEBUG, "%s: Endpoint selection strategy set to \"round-robin, starting from the top of the list\"\n", logstr);
            break;
        case ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START  :
            lcmaps_log(LOG_DEBUG, "%s: Endpoint selection strategy set to \"round-robin with random starting point in the list\"\n", logstr);
            break;
        case ENDPOINT_SELECTION_RANDOM                    :
            lcmaps_log(LOG_DEBUG, "%s: Endpoint selection strategy set to \"random endpoint selection\"\n", logstr);
            break;
	default:
	    lcmaps_log(LOG_ERR,
		"%s: Internal error, invalid Endpoint selection strategy %d\n",
		logstr,endpoint_selection_strategy);
	    return LCMAPS_MOD_FAIL;
    }


    lcmaps_log(LOG_DEBUG, "%s: Initialization succeeded\n", logstr);

    return LCMAPS_MOD_SUCCESS;
}


/******************************************************************************
Function:   plugin_introspect
Description:
    return list of required arguments
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_introspect(int *argc, lcmaps_argument_t **argv)
{
  const char *logstr = "lcmaps_plugin_c_pep-plugin_introspect()";
  static lcmaps_argument_t argList[] = {
    {"pem_string" , "char *"            ,  0, NULL},
    {"px509_chain", "STACK_OF(X509) *"  ,  0, NULL},
    {NULL         ,  NULL               , -1, NULL}
  };

  lcmaps_log(LOG_DEBUG, "%s: introspecting\n", logstr);

  *argv = argList;
  *argc = lcmaps_cntArgs(argList);
  lcmaps_log(LOG_DEBUG, "%s: address first argument: %p\n", logstr, (void*)argList);

  lcmaps_log(LOG_DEBUG, "%s: Introspect succeeded\n", logstr);

  return LCMAPS_MOD_SUCCESS;
}




/******************************************************************************
Function:   plugin_run
Description:
    Gather credentials for LCMAPS
Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_run(int argc, lcmaps_argument_t * argv)
{
    return plugin_run_or_verify(argc, argv, PLUGIN_RUN);
}


/******************************************************************************
Function:   plugin_verify
Description:
    Verify if user is entitled to use local credentials based on his grid
    credentials. This means that the site should already have been set up
    by, e.g., LCMAPS in a previous run. This method will not try to setup
    account leases, modify (distributed) passwd/group files, etc. etc.
    The outcome should be identical to that of plugin_run().
    In this particular case "plugin_verify()" is identical to "plugin_run()"

Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_verify(int argc, lcmaps_argument_t * argv)
{
  return plugin_run_or_verify(argc, argv, PLUGIN_VERIFY);
}

static int plugin_run_or_verify(int argc, lcmaps_argument_t *argv, int lcmaps_mode)
{
    const char *        logstr = NULL;
    char *              dnshostname = NULL;
    char *              pem_string = NULL;
    char *              clean_pem_string = NULL;
    char *              proxy_file_pilot = NULL;
    char *              pem_string_pilot = NULL;
    char *              clean_pem_string_pilot = NULL;
    void *              lcmaps_dummy_arg_val = NULL;
    long		j;
    unsigned int        u, random_choice = 0, selected_endpoint = 0;
    int                 should_do_high_level_retry;
    size_t              i = 0;
    pep_error_t         pepc_rc = 0;
    int                 rc = 0, downgrade_rc = 0, mod_rc = LCMAPS_MOD_FAIL;
    STACK_OF(X509)  *   px509_chain = NULL;
    /* if stored_real_uid==0, upgradeEffectiveToRealUid() does not do anything:
     * good as an initial value. */
    uid_t               stored_real_uid=0;
    uid_t               stored_eff_uid=0;

    pep_obligationhandler_t *oh=NULL;
    size_t		oh_count=0;

#if PEP_API != 1
    PEP *               pep_handle=NULL;
#endif

    if (lcmaps_mode == PLUGIN_RUN)
        logstr = "lcmaps_plugin_c_pep-plugin_run()";
    else if (lcmaps_mode == PLUGIN_VERIFY)
        logstr = "lcmaps_plugin_c_pep-plugin_verify()";
    else {
        lcmaps_log (LOG_ERR, "lcmaps_plugin_c_pep-plugin_run_or_verify(): attempt to run plugin in invalid mode: %d\n", lcmaps_mode);

        goto cleanup;
    }


    /* Grab PEM string when available */
    lcmaps_dummy_arg_val = lcmaps_getArgValue("pem_string", "char *", argc, argv);
    if (lcmaps_dummy_arg_val && (pem_string = *(char **) lcmaps_dummy_arg_val)) {
        lcmaps_log(LOG_DEBUG, "%s: found PEM string\n", logstr);
    } else {
        /* Fetch the X509 chain from the LCMAPS framework */
        lcmaps_dummy_arg_val = lcmaps_getArgValue("px509_chain", "STACK_OF(X509) *", argc, argv);
        if (lcmaps_dummy_arg_val && (px509_chain = *(STACK_OF(X509) **) lcmaps_dummy_arg_val)) {
            lcmaps_log(LOG_DEBUG,"%s: found X.509 chain.\n", logstr);
        } else {
            lcmaps_log (LOG_ERR,"%s: could not get the value of an X.509 chain or a PEM string.\n", logstr);
            goto cleanup;
        }

        /** Making a PEM string on the fly for Argus **/
        pem_string = lcmaps_stackofx509_to_pem(px509_chain);
        if (pem_string == NULL) {
            lcmaps_log (LOG_ERR, "%s: the certificate chain offered by the framework could not be converted to a PEM string\n", logstr);
            goto cleanup;
        }
    }

    /* the PEM string may contain a private key, which is used in verify_proxy
     * and friends. The P*P AuthZ framework, however, does not need the private
     * key. Remove it here to prevent needless exposure of the key. Don't mangle
     * the incoming string since it may be a static copy from the framework
     * structure.*/
    pem_string = lcmaps_plugin_c_pep_cleanse_pemstring(pem_string,&clean_pem_string);
    if ( !pem_string ) {
        lcmaps_log (LOG_ERR, "%s: could not remove the private key from the read proxyfile (of the user)\n", logstr);
        goto cleanup;
    }

    /* Do work here */

    /* Only AuthZ Interop profile uses the dnsname */
    if (use_profile & PROFILE_AUTHZ_INTEROP)	{
	lcmaps_log(LOG_DEBUG, "%s: Fetching DNS/Host name\n", logstr);
      #ifndef MAXHOSTNAMELEN
	   #define MAXHOSTNAMELEN ((size_t)255)
      #endif /* MAXHOSTNAMELEN */
	if ( (dnshostname = (char *) malloc (sizeof(char) * MAXHOSTNAMELEN)) == NULL)
	    goto cleanup;

	gethostname (dnshostname, MAXHOSTNAMELEN);
	lcmaps_log(LOG_DEBUG, "%s: DNS/Host name is: %s\n", logstr, dnshostname);
    }

    lcmaps_log(LOG_DEBUG, "PEP-C: initialize()\n");
#if PEP_API == 1
    if (pepc_initialize (&oh, &oh_count))
#else
    if ((pep_handle = pepc_initialize (&oh, &oh_count))==NULL)
#endif
    {
        lcmaps_log (LOG_ERR, "Error: Failed to initialize the PEP-C library\n");
        goto cleanup;
    }

    /* Fetch the pilot's proxy file (if needed) for client side authentication
     * of the SSL */
    if (PEP_OPTION_CLIENT_CERT_IMPLICIT == pep_option_client_cert_mode)	{
	/* Set proxy file location */
	proxy_file_pilot = getenv("X509_USER_PROXY");

	/* If proxy location is set and we are using the interop profile and if
	 * we want it send it is to be put into the SAML2-XACML2 environment */
	if ( proxy_file_pilot!=NULL &&
	     (use_profile & PROFILE_AUTHZ_INTEROP) &&
	     pep_option_send_pilot_cert_chain )
	{
	    /* Read from file (with a NFS rootsquash work-around) */
	    if ( !(pem_string_pilot = readProxyFile (proxy_file_pilot)) )   {
		lcmaps_log(LOG_ERR, "%s: could not read pilot proxy file\n",
			logstr);
		goto cleanup;
	    }

	    /* the PEM string may contain a private key, which is used in
	     * verify_proxy and friends. The P*P AuthZ framework, however, does
	     * not need the private key. Remove it here to prevent needless
	     * exposure of the key. Don't mangle the incoming string since it
	     * may be a static copy from the framework structure.
	     */
	    if (!(lcmaps_plugin_c_pep_cleanse_pemstring(pem_string_pilot, &clean_pem_string_pilot)))
	    {
		lcmaps_log (LOG_ERR, "%s: could not remove the private key from the read proxyfile (of the Pilot)\n", logstr);
		goto cleanup;
	    }
	}
    }

    lcmaps_log(LOG_DEBUG, "PEP-C: construct_request()\n");
    if (pepc_construct_request (
        use_profile,
        pem_string,
        actiontype,
        actionid,
        resourceid,
        resourcetype,
        dnshostname,
        clean_pem_string_pilot))
    {
        lcmaps_log (LOG_ERR, "Error: Failed to construct the Request\n");
        goto cleanup;
    }


#if PEP_API == 1
    /* Set pep options */

    /* Setup the PEP-API loglevel, the piped log output for cURL and callback
     * for the rest */
    if (pepapi_setup_log(pep_c_loglevel))
	goto cleanup;

    /* Set SSL cipher list, by default to be compatible with OpenSSL 1.0
     * client-side.  This setting is only available in pep-api 2.0. Setting it
     * here is a workaround, since this really should be fixed on server-side
     */
    /* Change: The cipherlist strings for libnss, openssl, gnutls and others
     * are different. The autotools are leading the define. By default it's an
     * empty string "" to ensure that our options are leading. This is tested
     * on CentOS5 (openssl) and CentOS6 (libnss).
     */
    /* Change2: Only set a cipher list when explicitly ./configure'd or
     * configured in the lcmaps.db file.
     */
    if (pep_option_ssl_cipher_list) {
        pepc_rc=pep_setoption(PEP_OPTION_ENDPOINT_SSL_CIPHER_LIST, pep_option_ssl_cipher_list);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(PEP_OPTION_ENDPOINT_SSL_CIPHER_LIST,\"%s\") failed: %d: %s\n",
			pep_option_ssl_cipher_list, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
    }

    /* PEP_OPTION_ENDPOINT_SSL_VALIDATION */
    if ( pep_option_ssl_validation != PEP_OPTION_SSL_VALIDATION_DEFAULT ) {
        if ( pep_option_ssl_validation == PEP_OPTION_SSL_VALIDATION_ON ) {
            pepc_rc = pep_setoption(PEP_OPTION_ENDPOINT_SSL_VALIDATION, 1);
            if (pepc_rc != PEP_OK) {
                lcmaps_log (LOG_ERR,
			    "Error: pep_setoption(PEP_OPTION_ENDPOINT_SSL_VALIDATION,1) failed: %d: %s\n",
                            (int)pepc_rc, pep_strerror(pepc_rc));
                goto cleanup;
            }
        } else if ( pep_option_ssl_validation == PEP_OPTION_SSL_VALIDATION_OFF ) {
            pepc_rc = pep_setoption(PEP_OPTION_ENDPOINT_SSL_VALIDATION, 0);
            if (pepc_rc != PEP_OK) {
                lcmaps_log (LOG_ERR,
			    "Error: pep_setoption(PEP_OPTION_ENDPOINT_SSL_VALIDATION,0) failed: %d: %s\n",
                            (int)pepc_rc, pep_strerror(pepc_rc));
                goto cleanup;
            }
        } else {
            lcmaps_log (LOG_ERR, "LCMAPS internal error: pep_option_ssl_validation set to invalid value (%d)\n",
                        pep_option_ssl_validation);
            goto cleanup;
        }
    }

    /* PEP_OPTION_ENDPOINT_CLIENT_CERT and PEP_OPTION_ENDPOINT_CLIENT_KEY */
    if ( PEP_OPTION_CLIENT_CERT_EXPLICIT == pep_option_client_cert_mode ) {
        if ( !pep_option_client_cert || !pep_option_client_key ) {
            lcmaps_log (LOG_ERR, "Error: assert failed: setoptions for explicit client cert and key needs both set.\n");
            goto cleanup;
        }
        pepc_rc = pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_CERT, pep_option_client_cert);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
		    "Error: pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_CERT,%s) failed: %d: %s\n",
                    pep_option_client_cert, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
        pepc_rc = pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_KEY, pep_option_client_key);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_KEY,%s) failed: %d: %s\n",
                        pep_option_client_key, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
    } else if ( PEP_OPTION_CLIENT_CERT_IMPLICIT == pep_option_client_cert_mode ) {
        if ( !proxy_file_pilot ) {
            lcmaps_log (LOG_ERR, "Error: assert failed: setoptions for implicit ssl needs X509_USER_PROXY set.\n");
            goto cleanup;
        }
        /* This is not the case anymore as the PEPd will change it behavior in a future release of Argus */
        /* lcmaps_log (LOG_ERR, "Warning: implicit SSL client auth with invokers proxy will not influence authorization decision by PEPd.\n"); */
        pepc_rc = pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_CERT, proxy_file_pilot);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_CERT,%s) failed: %d: %s\n",
                        proxy_file_pilot, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
        pepc_rc = pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_KEY, proxy_file_pilot);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_KEY,%s) failed: %d: %s\n",
                        proxy_file_pilot, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
    } else if ( PEP_OPTION_CLIENT_CERT_NONE == pep_option_client_cert_mode ) {
        lcmaps_log(LOG_DEBUG, "pep_option_client_cert_mode set to none\n");
    } else {
        lcmaps_log (LOG_ERR, "LCMAPS internal error: pep_option_client_cert_mode set to invalid value  at this point (%d)\n",
                    pep_option_ssl_validation);
        goto cleanup;
    }

    /* PEP_OPTION_ENDPOINT_CLIENT_KEYPASSWORD */
    if ( pep_option_client_keypassword ) {
        pepc_rc = pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_KEYPASSWORD, pep_option_client_keypassword);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_KEYPASSWORD,****SUPPRESSED****) failed: %d: %s\n",
                        (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
    }

    /* PEP_OPTION_ENDPOINT_SERVER_CAPATH */
    if ( pep_option_server_capath ) {
        pepc_rc = pep_setoption(PEP_OPTION_ENDPOINT_SERVER_CAPATH, pep_option_server_capath);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(PEP_OPTION_ENDPOINT_SERVER_CAPATH,%s) failed: %d: %s\n",
                        pep_option_server_capath, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
    }

    /* PEP_OPTION_ENDPOINT_SERVER_CERT */
    if ( pep_option_server_cert ) {
        pepc_rc = pep_setoption(PEP_OPTION_ENDPOINT_SERVER_CERT, pep_option_server_cert);
        if (pepc_rc != PEP_OK)
        {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(PEP_OPTION_ENDPOINT_SERVER_CERT,%s) failed: %d: %s\n",
                        pep_option_server_cert, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
    }
    
    /* PEP_OPTION_ENDPOINT_SERVER_CERT via --use-pilot-proxy-as-cafile */
    if ( pep_option_use_pilot_proxy_as_cafile ) {
        pepc_rc = pep_setoption(PEP_OPTION_ENDPOINT_SERVER_CERT, proxy_file_pilot);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(PEP_OPTION_ENDPOINT_SERVER_CERT,%s) failed: %d: %s\n",
                        proxy_file_pilot, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
    }
/* PEP_API == 1 */
#else
/* PEP_API != 1 */
    /* Set pep options */

    /* Setup the PEP-API loglevel, the piped log output for cURL and callback
     * for the rest */
    if (pepapi_setup_log(pep_handle, pep_c_loglevel))
	goto cleanup;

    /* Set SSL cipher list, by default to be compatible with OpenSSL 1.0
     * client-side.  This setting is only available in pep-api 2.0. Setting it
     * here is a workaround, since this really should be fixed on server-side
     */
    /* Change: The cipherlist strings for libnss, openssl, gnutls and others
     * are different. The autotools are leading the define. By default it's an
     * empty string "" to ensure that our options are leading. This is tested
     * on CentOS5 (openssl) and CentOS6 (libnss).
     */
    /* Change2: Only set a cipher list when explicitly ./configure'd or
     * configured in the lcmaps.db file.
     */
    if (pep_option_ssl_cipher_list) {
        pepc_rc=pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SSL_CIPHER_LIST, pep_option_ssl_cipher_list);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SSL_CIPHER_LIST,\"%s\") failed: %d: %s\n",
			pep_option_ssl_cipher_list, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
    }

    /* PEP_OPTION_ENDPOINT_SSL_VALIDATION */
    if ( pep_option_ssl_validation != PEP_OPTION_SSL_VALIDATION_DEFAULT ) {
        if ( pep_option_ssl_validation == PEP_OPTION_SSL_VALIDATION_ON ) {
            pepc_rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SSL_VALIDATION, 1);
            if (pepc_rc != PEP_OK) {
                lcmaps_log (LOG_ERR,
			    "Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SSL_VALIDATION,1) failed: %d: %s\n",
                            (int)pepc_rc, pep_strerror(pepc_rc));
                goto cleanup;
            }
        } else if ( pep_option_ssl_validation == PEP_OPTION_SSL_VALIDATION_OFF ) {
            pepc_rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SSL_VALIDATION, 0);
            if (pepc_rc != PEP_OK) {
                lcmaps_log (LOG_ERR,
			    "Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SSL_VALIDATION,0) failed: %d: %s\n",
                            (int)pepc_rc, pep_strerror(pepc_rc));
                goto cleanup;
            }
        } else {
            lcmaps_log (LOG_ERR, "LCMAPS internal error: pep_option_ssl_validation set to invalid value (%d)\n",
                        pep_option_ssl_validation);
            goto cleanup;
        }

    }

    /* PEP_OPTION_ENDPOINT_CLIENT_CERT and PEP_OPTION_ENDPOINT_CLIENT_KEY */
    if ( PEP_OPTION_CLIENT_CERT_EXPLICIT == pep_option_client_cert_mode ) {
        if ( !pep_option_client_cert || !pep_option_client_key ) {
            lcmaps_log (LOG_ERR, "Error: assert failed: setoptions for explicit client cert and key needs both set.\n");
            goto cleanup;
        }
        pepc_rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_CERT, pep_option_client_cert);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_CERT,%s) failed: %d: %s\n",
                        pep_option_client_cert, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
        pepc_rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_KEY, pep_option_client_key);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_KEY,%s) failed: %d: %s\n",
                        pep_option_client_key, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
    } else if ( PEP_OPTION_CLIENT_CERT_IMPLICIT == pep_option_client_cert_mode ) {
        if ( !proxy_file_pilot ) {
            lcmaps_log (LOG_ERR, "Error: assert failed: setoptions for implicit ssl needs X509_USER_PROXY set.\n");
            goto cleanup;
        }
        /* This is not the case anymore as the PEPd will change it behavior in a future release of Argus */
        /* lcmaps_log (LOG_ERR, "Warning: implicit SSL client auth with invokers proxy will not influence authorization decision by PEPd.\n"); */
        pepc_rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_CERT, proxy_file_pilot);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_CERT,%s) failed: %d: %s\n",
                        proxy_file_pilot, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
        pepc_rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_KEY, proxy_file_pilot);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_KEY,%s) failed: %d: %s\n",
                        proxy_file_pilot, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
    } else if ( PEP_OPTION_CLIENT_CERT_NONE == pep_option_client_cert_mode ) {
        lcmaps_log(LOG_DEBUG, "pep_option_client_cert_mode set to none\n");
    } else {
        lcmaps_log (LOG_ERR, "LCMAPS internal error: pep_option_client_cert_mode set to invalid value  at this point (%d)\n",
                    pep_option_ssl_validation);
        goto cleanup;
    }

    /* PEP_OPTION_ENDPOINT_CLIENT_KEYPASSWORD */
    if ( pep_option_client_keypassword ) {
        pepc_rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_KEYPASSWORD, pep_option_client_keypassword);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_KEYPASSWORD,****SUPPRESSED****) failed: %d: %s\n",
                        (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
    }

    /* PEP_OPTION_ENDPOINT_SERVER_CAPATH */
    if ( pep_option_server_capath ) {
        pepc_rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SERVER_CAPATH, pep_option_server_capath);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SERVER_CAPATH,%s) failed: %d: %s\n",
                        pep_option_server_capath, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
    }

    /* PEP_OPTION_ENDPOINT_SERVER_CERT */
    if ( pep_option_server_cert ) {
        pepc_rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SERVER_CERT, pep_option_server_cert);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SERVER_CERT,%s) failed: %d: %s\n",
                        pep_option_server_cert, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
    }
    
    /* PEP_OPTION_ENDPOINT_SERVER_CERT via --use-proxy-as-cafile */
    if ( pep_option_use_pilot_proxy_as_cafile ) {
        pepc_rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SERVER_CERT, proxy_file_pilot);
        if (pepc_rc != PEP_OK) {
            lcmaps_log (LOG_ERR,
			"Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SERVER_CERT,%s) failed: %d: %s\n",
                        proxy_file_pilot, (int)pepc_rc, pep_strerror(pepc_rc));
            goto cleanup;
        }
    }
#endif /* PEP_API != 2 */


    lcmaps_log(LOG_DEBUG, "PEP-C: engage()\n");

    /* Downgrade effective privileges to cope with NFS mounted file systems with root squashing */
    if ( (downgrade_rc=downgradeEffectiveToRealUid (&stored_real_uid, &stored_eff_uid)) )    {
	lcmaps_log(LOG_ERR,"%s: Error: cannot downgrade to real uid\n", logstr);
	goto cleanup;
    }

    /* retry unless otherwise specified */
    should_do_high_level_retry=1;

    switch (endpoint_selection_strategy)
    {
	/* Round Robin - Top to Bottom */
        case ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM :
            lcmaps_log(LOG_DEBUG, "%s: Starting Round-Robin (top to bottom) strategy\n", logstr);

            for (j = 0; j < endpoint_retrycount && should_do_high_level_retry; j++) {
		for (u = 0; u < n_endpoints && should_do_high_level_retry; u++) {
		    lcmaps_log(LOG_INFO, "%s: Using endpoint %s, try #%lu\n",
			    logstr, pep_daemon_endpoint_urls[u], (long unsigned)j+1);

		    /* Do actual callout */
#if PEP_API == 1
		    rc=pepc_engage (pep_daemon_endpoint_urls[u],
					 n_endpoints,
					 &should_do_high_level_retry);
#else
		    rc=pepc_engage (pep_handle,
					 pep_daemon_endpoint_urls[u],
					 n_endpoints,
					 &should_do_high_level_retry);
#endif
                } /* endpoint */
            } /* try */
            break; /* switch endpoint_selection_strategy */
	case ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START :
	    lcmaps_log(LOG_DEBUG, "%s: Starting Round-Robin (random start) strategy\n", logstr);
            
	    /* Pick a random starting point */
            random_choice = pepc_rand() % n_endpoints;

            for (j = 0; j < endpoint_retrycount && should_do_high_level_retry; j++) {
		for (u = 0; u < n_endpoints && should_do_high_level_retry; u++) {
		    selected_endpoint = (random_choice + u) % n_endpoints;

		    lcmaps_log(LOG_INFO, "%s: Using endpoint %s, try #%lu\n",
			    logstr, pep_daemon_endpoint_urls[selected_endpoint], (long unsigned)j+1);

		    /* Do actual callout */
#if PEP_API == 1
		    rc=pepc_engage (pep_daemon_endpoint_urls[selected_endpoint],
					 n_endpoints,
					 &should_do_high_level_retry);
#else
		    rc=pepc_engage (pep_handle,
					 pep_daemon_endpoint_urls[selected_endpoint],
					 n_endpoints,
					 &should_do_high_level_retry);
#endif
                } /* endpoint */
            } /* try */
            break; /* switch endpoint_selection_strategy */
    case ENDPOINT_SELECTION_RANDOM :
            lcmaps_log(LOG_DEBUG, "%s: Starting Random selection strategy\n", logstr);

            for (j = 0; j < endpoint_retrycount && should_do_high_level_retry; j++) {
		for (u = 0; u < n_endpoints && should_do_high_level_retry; u++) {
		    /* Pick a random choice */
		    random_choice = pepc_rand() % n_endpoints;

		    lcmaps_log(LOG_INFO, "%s: Using endpoint %s\n",
			    logstr, pep_daemon_endpoint_urls[random_choice]);

		    /* Do actual callout */
#if PEP_API == 1
		    rc=pepc_engage (pep_daemon_endpoint_urls[random_choice],
					 n_endpoints,
					 &should_do_high_level_retry);
#else
		    rc=pepc_engage (pep_handle,
					 pep_daemon_endpoint_urls[random_choice],
					 n_endpoints,
					 &should_do_high_level_retry);
#endif
                } /* endpoint */
            } /* try */
            break; /* switch endpoint_selection_strategy */
	default:
	    lcmaps_log(LOG_ERR,"%s: unknown endpoint_selection_strategy %u\n",
		    __func__, endpoint_selection_strategy);
	    goto cleanup;
    }

    /* If pepc_engage succeeded -> plugin succeeded */
    if (rc==0)
	mod_rc=LCMAPS_MOD_SUCCESS;

cleanup:
    lcmaps_log(LOG_DEBUG, "PEP-C: cleanup()\n");
#if PEP_API == 1
    if (pepc_cleanup ())
#else
    if (pepc_cleanup (pep_handle))
#endif
        lcmaps_log (LOG_ERR, "Error: Failed to clean up PEP-C.\n");

    /* Restore privileges to previous state when we successfully downgraded */
    if (downgrade_rc==0 && upgradeEffectiveToRealUid (stored_real_uid, stored_eff_uid))
	/* Upgrade failed -> plugin failed */
	mod_rc=LCMAPS_MOD_FAIL;

    /* Clean up all memory */
    free (dnshostname);
    free (clean_pem_string);
    free (pem_string_pilot);
    free (clean_pem_string_pilot);

    for (i=0; i<oh_count; i++) free(oh[i].id);
    free (oh);

    /* Clean up supported obligations, not needed after registration */
    free_supported_obligations_list();

    /* Log result of plugin */
    lcmaps_log(LOG_INFO, "%s: c_pep plugin %s\n",
	    logstr, mod_rc==LCMAPS_MOD_SUCCESS ? "succeeded" : "failed");

    /* Return status */
    return mod_rc;
}



/******************************************************************************
Function:   plugin_terminate
Description:
    Terminate plugin
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_terminate()
{
    const char * logstr = "lcmaps_plugin_c_pep-plugin_terminate()";
    lcmaps_log(LOG_DEBUG,"%s: terminating\n", logstr);

    /* return any memory allowed by subsidiaries */
    if ( pep_option_server_capath_default )
        free( pep_option_server_capath_default );

    return LCMAPS_MOD_SUCCESS;
}


/******************************************************************************
Function:   plugin_c_pep_readpwfile
Description:
    Read a password (single line) from a file
Parameters:
    static char *f, char *staticbuf, int n;
Returns:
    pointer to the buffer with the password or NULL on failure
******************************************************************************/

static char *plugin_c_pep_readpwfile(char *fn, char *buf, int n)
{
  FILE *f;
  char *s;

  if ( !fn ) {
      lcmaps_log (LOG_ERR, "plugin_c_pep_readpwfile: called with an empty file name.\n");
      return NULL;
  }
  if ( ( n <= 0 ) || !buf ) {
      lcmaps_log (LOG_ERR, "plugin_c_pep_readpwfile: string buffer size zero or negative or buffer is a NULL pointer.\n");
      return NULL;
  }

  if ( !(f = fopen(fn,"r")) ) {
      lcmaps_log (LOG_ERR, "plugin_c_pep_readpwfile: cannot open password file %s.\n",fn);
      return NULL;
  }
  errno=0;
  s = fgets(buf,n,f);
  if ( !s ) {
    if ( errno ) { /* a true error condition */
      lcmaps_log (LOG_ERR, "plugin_c_pep_readpwfile: cannot read password file %s: error %d.\n", fn,errno);
      fclose(f);
      return NULL;
    } else {
      /* EOF reached, password is empty file */
      *buf = '\0';
    }
  }
  fclose(f);

  if ( strlen(buf) == (size_t)n ) {
      lcmaps_log (LOG_ERR, "password in file %s is too long (at or over %d characters.\n", fn,n-1);
      return NULL;
  }

  /* strip new-line or CR characters for any platform */
  s = buf+strlen(buf);
  while ( ( s > buf ) && ('\n' == *s || '\r' == *s) ) { *s = '\0'; s--; }

  return buf;
}


/******************************************************************************
Function:   plugin_c_pep_optequal(s,t)
Description:
    See if <s> is the option "-[-]<t>"
Parameters:
    const char *s, const char *t
Returns:
    0	if <s> is NOT this option
    1	if <s> IS this option
    -1	on error
******************************************************************************/
static int plugin_c_pep_optequal(const char *s, const char *t)
{
	register size_t n,m;
	const char *p = s;

	if ( !s || !t ) {
		lcmaps_log (LOG_ERR,"plugin_c_pep_optequal: invalid invocation with one or two NULL pointers.\n");
		return -1;
	}

	n = strlen(s);
	m = strlen(t);

	/* it should be an option and start with at least one dash */
	if ('-' != *p ) return 0;

	/* eat at most two dashes from p */
	if ('-' == *p) p++;
	if ('-' == *p) p++;

	n = strlen(p);
	if ( n != m ) return 0;

        if ( strncasecmp(p, t, m) == 0 ) return 1;
	return 0;
}


/******************************************************************************
Function:   lcmaps_plugin_c_pep_cleanse_pemstring(char *in, char **path)
Description:
    Remove any data but certs from a PEM string and store the result in
    a newly-alloced buffer. Return a pointer to the buffer as well, or
    NULL on failure. In case of failure, no malloc will persist.
******************************************************************************/

static char *lcmaps_plugin_c_pep_cleanse_pemstring(char *in, char **destp)
{
    char *s,*t,*p;
    size_t  i;
    static const char *pem_begin_certificate = "-----BEGIN CERTIFICATE-----";
    static const char *pem_end_certificate = "-----END CERTIFICATE-----";

    if ( !in || !destp ) {
        return NULL;
    }
    s = malloc(strlen(in)+1);
    if ( !s ) {
      return NULL;
    }

    t = in;
    p = s;
    while ( *t ) {
        if ( '-' == *t ) {
            if ( strncmp(t,pem_begin_certificate,strlen(pem_begin_certificate)) == 0 ) {
                /* a cert blob is starting here, copy until the end of the cert */
                for( i = 0; *t && (i < strlen(pem_begin_certificate)); i++ ) *(p++) = *(t++);
                while ( 1 ) {
                    if ( ! *t ) break;
                    if ( '-' == *t && ( strncmp(t,pem_end_certificate,strlen(pem_end_certificate)) == 0 ) ) {
                        for( i = 0; *t && (i < strlen(pem_end_certificate)); i++ ) *(p++) = *(t++);
                        /* include any training new line, usually one has at least one, then get out */
                        while ( '\r' == *t || '\n' == *t ) *(p++) = *(t++);
                        break;
                    }
                    *(p++) = *(t++);
                }
            } else
                t++;
        } else
            t++;
    }
    *p = '\0';

    *destp = s;
    return s;
}


/******************************************************************************
Function:   plugin_c_pep_set_capath_default(char **path)
Description:
    Determine the default CApath directory, if there is any. It will look
    for X509_CERT_DIR, then for /etc/grid-security/certificates
Parameters:
    address of the pointer to string to be filled with the directory name
Returns:
    the name of the directory, pointing to the piece of memory alloc'ed by
    this function
    NULL if no default directory was found (no memory will have been alloced and
    *path remains unmodified)
******************************************************************************/

static char *plugin_c_pep_set_capath_default(char **p)
{
    char  *dirname;
    int   rc;
    size_t len;
    struct stat sbuf;
    const char  *targets[3];
    int   i;
    char  tmpname[PATH_MAX],*homedir=NULL;
    uid_t ruid,euid;
#ifdef HAVE_GETRESUID
    uid_t suid;
#endif
    /* Don't use .globus/certificates unless we checked we want to */
    int use_dotglobus=0;

    if ( !p ) {
        lcmaps_log (LOG_ERR,"%s: invalid invocation: address to a pointer to buffer is nil.\n", __func__);
        return NULL;
    }

    /* build the search path list, ensure that X509_CERT_DIR is at the end
       since getenv(3) semantics define the output buffer to be volatile,
       i.e. a direct result from getenv(3) may only be in the list once */
    targets[0] = getenv("X509_CERT_DIR");
    targets[2] = "/etc/grid-security/certificates";

#ifdef HAVE_GETRESUID
    /* Get real/effective/saved uid */
    if (getresuid(&ruid,&euid,&suid))	{
	lcmaps_log (LOG_ERR,
		"%s: Error calling getresuid(): %s\n",__func__,strerror(errno));
	return NULL;
    } else  {
	/* Is real==effective==saved != 0 ? */
	if (ruid==euid && ruid==suid && ruid!=0)
	    use_dotglobus=1;
    }
#else
    /* Get real/effective uid */
    ruid=getuid();
    euid=geteuid();
    /* See if real==effective != 0 */
    if (ruid==euid && ruid!=0)	{
	/* Check if we have a saved uid 0 */
	if (seteuid(0)!=0 && errno==EPERM)
	    use_dotglobus=1; /* seteuid failed */
	else	{
	    /* revert to previous euid */
	    if (seteuid(euid))	{ /* revert failed ?! */
		lcmaps_log(LOG_ERR,
			"%s: Cannot revert effective uid to %d: %s\n",
			__func__, euid, strerror(errno));
		 return NULL;
	    }
	}
    }
#endif

    /* Whether we can include a dotglobus */ 
    if (use_dotglobus &&
	( homedir=getenv("HOME") ) &&
	( strlen(homedir) < (PATH_MAX-21)) )
    {
	sprintf(tmpname,"%s/.globus/certificates",homedir);
	targets[1] = tmpname;
    } else  {
	targets[1] = NULL;
    }

    for ( i = 0 ; i < 3 ; i++ ) {
        if ( targets[i] ) {
            len = strlen(targets[i]);
            rc = stat(targets[i],&sbuf);
            if ( rc ) { /* does not exist or any other error, don't use this */
                continue;
            }
            if ( !S_ISDIR(sbuf.st_mode) ) {
                lcmaps_log (LOG_ERR,
			"%s: %s exists but is not a directory, ignored\n",
			__func__, targets[i]);
                continue;
            }
            len = strlen(targets[i]);
            dirname = malloc(len+1);
            if ( !dirname ) {
                lcmaps_log (LOG_ERR,
			"%s: out of memory when mallocing space for copy of default CA dir %s\n",
			__func__, targets[i]);
                *p = NULL;
                return NULL;
            }
            strncpy(dirname,targets[i],len);
            dirname[len+1] = '\0';
            *p = dirname;
            return dirname;
        }
    }

    return NULL;
}


static char * readProxyFile (const char * proxyfile)
{
    int                 fd = -1;
    char *              pem_buf = NULL;
    ssize_t             rc = 0;
    struct stat         s1, s2;

    uid_t stored_real_uid;
    uid_t stored_eff_uid;

    /* No input, nothing to do */
    if (proxyfile == NULL) {
        lcmaps_log (LOG_ERR,
		"%s: No proxy file to read from (filename is NULL).\n",
		__func__);
        return NULL;
    }

    /* Downgrade effective privileges to cope with NFS mounted file systems with root squashing */
    if (downgradeEffectiveToRealUid (&stored_real_uid, &stored_eff_uid))
	return NULL;

    /* Read PEM string */
    if ((fd = open (proxyfile, O_RDONLY)) < 0) {
        lcmaps_log (LOG_ERR,
		"%s: Could not open file for reading \"%s\": %s\n",
		__func__, proxyfile, strerror(errno));
	goto cleanup;
    }
        
    /* Stat the proxy file */
    if (fstat(fd, &s1) < 0) {
	lcmaps_log (LOG_ERR,
		"%s: Could not stat opened file \"%s\": %s\n",
		__func__, proxyfile, strerror(errno));
	goto cleanup;
    }
	
    /* Allocate buffer with the size of the opened file */
    pem_buf = (char *) calloc ((size_t)1, sizeof (char) * (size_t)(s1.st_size) + 1);
    if (pem_buf == NULL)    {
	lcmaps_log(LOG_ERR, "%s: Out of memory\n", __func__);
	goto cleanup;
    }

    /* Read PEM into the buffer */
    rc = read(fd, pem_buf, (size_t)(s1.st_size));
    /* Check for errors */
    if (rc < 0) {
	lcmaps_log (LOG_ERR,
		"%s: Read error on file \"%s\": %s.\n",
		__func__, proxyfile, strerror(errno));
	free (pem_buf); pem_buf = NULL;
	goto cleanup;
    }
    /* Check for unexpected file size */
    if (rc!=(ssize_t)(s1.st_size))  {
	lcmaps_log(LOG_ERR,
		"%s: read() returned %ld instead of expected %ld\n",
		__func__, (long)rc, (long)(s1.st_size));
	free (pem_buf); pem_buf = NULL;
	goto cleanup;
    }

    /* Perform safety check to notice tampering with the file */
	
    /* Stat the proxy file */
    if (fstat (fd, &s2) < 0) {
	lcmaps_log (LOG_ERR,
		"%s: Could not re-stat opened file \"%s\". Reason: %s\n",
		__func__, proxyfile, strerror(errno));
	free (pem_buf); pem_buf = NULL;
	goto cleanup;
    }
    
    /* Safety check, simple check to see if the file has not been tampered with during the read. */
    if ((s1.st_dev == s2.st_dev) &&
	(s1.st_ino == s2.st_ino) &&
	(s1.st_size == s2.st_size))
    {
	lcmaps_log(LOG_DEBUG,
		"%s: File \"%s\" was read safely.\n", __func__, proxyfile);
    }
    else
    {
	lcmaps_log(LOG_ERR,
		"%s: File \"%s\" has been moved or written to during the read.\n",
		__func__, proxyfile);
	free (pem_buf); pem_buf = NULL;
    }

cleanup:
    /* Close the file */
    if (fd>=0)
	close(fd);

    /* Restore privileges to previous state */
    if (upgradeEffectiveToRealUid (stored_real_uid, stored_eff_uid))	{
	/* failed after all */
	free(pem_buf); pem_buf = NULL;
    }

    /* Return the buffer (or NULL) */
    return pem_buf;
}


/*  When the proxy is located on an NFS mount and on the server side the root squash
 *  option has been enabled, the effective uid is mapped to user 'nobody' which should
 *  not be able to read the proxy file. To work around this problem, the effective
 *  uid of the process is changed to that of the calling user and once glexec is done,
 *  the saved uid is used to restore the identity of the process,
 */
static int downgradeEffectiveToRealUid (uid_t *real_uid, uid_t * saved_uid)
{
    /* When non-root, downgrade to real uid */
    if ( (*real_uid = getuid()) != 0 )	{ /* Save real uid */
        *saved_uid = geteuid();	/* Save effective uid */
        if (seteuid(*real_uid)) {
            lcmaps_log(LOG_ERR, "%s: seteuid() failed: %s\n",
		       __func__, strerror(errno));
            return 1;
        }
    }
    return 0;
}

static int upgradeEffectiveToRealUid (uid_t real_uid, uid_t saved_uid)
{
    /* When non-root, revert to previous uid */
    if (real_uid != 0) {
        if (seteuid(saved_uid)) {   /* revert */
            lcmaps_log(LOG_ERR, "%s: seteuid() failed: %s\n",
		       __func__, strerror(errno));
            return 1;
        }
    }
    return 0;
}

/******************************************************************************
Function:   lcmaps_stackofx509_to_pem
Description:
    Will create a PEM string encoded representation of the certificate chain
Parameters:
    A stack of X.509 certificates.
Returns:
    NULL : The input could not be transformed.
   !NULL : The input is encoded into a PEM string
******************************************************************************/
static char * lcmaps_stackofx509_to_pem (STACK_OF(X509) * px509_chain)
{
    int     i = 0;
    int     retval = 0;
    long    len = 0;
    BIO *   certbio_mem = NULL;
    X509 *  px509       = NULL;
    char *  pem_stringP  = NULL;
    char *  pem_string   = NULL;

    /* Initialize the BIO memory */
    certbio_mem = BIO_new(BIO_s_mem());

    /* Push certificates in chain into the BIO memory stack */
    for (i = 0; i < sk_X509_num(px509_chain); i++)
    {
        px509 = sk_X509_value(px509_chain, i);
        retval = PEM_write_bio_X509(certbio_mem, px509);
	if (retval==0 || retval==-1) /* See man pem(3ssl): 0 is failure, 1
					success */
	    return NULL;
    }

    /* Push BIO memory to character array */
    /* Warning: The BIO_get_mem_data() will only apply a REFERENCE to
     * the pem_string, NOT A COPY of the data */
    len = BIO_get_mem_data (certbio_mem, &pem_stringP);
    if (len<0) return NULL;

    pem_string = malloc((size_t)len + 1);
    if (pem_string) {
        strncpy(pem_string, pem_stringP, (size_t)len);
        pem_string[len]='\0';
    }

    /* Closing and Freeing buffers AFTER, we can ignore the return value, since
     * it's always 1 according to the man page, but this way we at least prevent
     * compiler and cppcheck warnings. */
    if (BIO_set_close(certbio_mem, BIO_CLOSE)!=1)   {
	lcmaps_log(LOG_ERR,
	    "%s: something is wrong: BIO_set_close() should always return 1\n",
	    __func__);
    }

    BIO_free(certbio_mem);

    /* Needs freeing, assuming the malloc() worked. */
    return pem_string;
}


