/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

/* Define both _XOPEN_SOURCE and _GNU_SOURCE. Latter for F_SETPIPE_SZ */
#define _XOPEN_SOURCE	500
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <syslog.h>

/* include lcmaps_c_pep_config.h for PEP_API */
#include "lcmaps_c_pep_config.h"

/* Includes for PEP-C library */
#if PEP_API == 1
#include <pep/pep.h>
#else
#include <argus/pep.h>
#endif

#include <lcmaps/lcmaps_log.h>

#include "pep-c-log.h"

/******************************************************************************
                                Definitions
******************************************************************************/

/* The next buffer is used for logging the cURL output */
#define PIPE_BUFFER_SIZE ((size_t)2048)

/* Used in case we have F_SETPIPE_SZ to increase the pipe size */
#ifdef F_SETPIPE_SZ
# define PIPE_SIZE   (1048576)
#endif

#define LOG_BUFFER_SIZE ((size_t)2048)

/******************************************************************************
                                Static globals
******************************************************************************/

static int pfd[2] = {-1,-1};/* Pipe file descriptors */
static FILE *f_err = NULL;  /* File pointer associated with write-end pipe */


/******************************************************************************
                            Local function prototypes
******************************************************************************/

/**
 * pep_log_handler_callback function, catches the PEP library logging and
 * forwards into the LCMAPS logging.
 * Parameters:
 *	level - PEP_LOGLEVEL_ERROR, PEP_LOGLEVEL_WARN, PEP_LOGLEVEL_INFO or
 *	        PEP_LOGLEVEL_DEBUG
 *	fmt   - format string
 *	args  - log arguments
 */
#ifdef __GNUC__
static int pepapi_log(int level, const char *fmt, va_list args)
    __attribute__((format(printf,2,0)));
#else
static int pepapi_log(int level, const char *fmt, va_list args);
#endif

/******************************************************************************
                            Function definitions
******************************************************************************/

/**
 * Setup logging for pep-api library, including the piped log output for cURL,
 * otherwise it will go to stderr. Actual log will be logged to lcmaps log by
 * pep_log_curl()
 * argument pep_handle (PEP_API>=2), used to get the curl handle
 *	    pep_c_loglevel, loglevel for pep-api-c-library.
 * Returns 0 on success, -1 on error
 */
#if PEP_API == 1
int pepapi_setup_log(int pep_c_loglevel)	{
#else
int pepapi_setup_log(PEP *pep_handle, int pep_c_loglevel)	{
#endif
    pep_error_t pep_rc;
    int saved_errno=0;
#if defined(F_SETPIPE_SZ) && defined(F_GETPIPE_SZ)
    int fcntl_rc;
#endif

    /* First create a pipe pair, stored in global static */
    if (pipe(pfd)) {
        lcmaps_log(LOG_ERR,
	    "%s: Error - cannot open new pipes: (errno=%d, %s)\n",
	    __func__, errno, strerror(errno));
        return -1;
    }

    /* Set FD_CLOEXEC flags on them */
    if ( (fcntl(pfd[1], F_SETFD, FD_CLOEXEC) == -1) ||
	 (fcntl(pfd[0], F_SETFD, FD_CLOEXEC) == -1) ) {
        lcmaps_log(LOG_ERR,
	    "%s: Error - cannot set pipes to CLOEXEC: (errno=%d, %s)\n",
	    __func__, errno, strerror(errno));
	goto cleanup;
    }
    /* Set non-blocking flags on them */
    if ( (fcntl(pfd[1], F_SETFL, O_NONBLOCK) == -1) ||
	 (fcntl(pfd[0], F_SETFL, O_NONBLOCK) == -1) ) {
        lcmaps_log(LOG_ERR,
	    "%s: Error - cannot set write pipe to non-blocking: (errno=%d, %s)\n",
	    __func__, errno, strerror(errno));
	goto cleanup;
    }
    /* If we have F_SETPIPE_SZ increase the pipe size since the default is
     * typically only 64kB */
#if defined(F_SETPIPE_SZ) && defined(F_GETPIPE_SZ)
    /* Get current size: only want to increase when too small */
    if ( (fcntl_rc=fcntl(pfd[1], F_GETPIPE_SZ))==-1 )	{
        lcmaps_log(LOG_ERR,
	    "%s: Error - cannot get current pipe-size: (errno=%d, %s)\n",
	    __func__, errno, strerror(errno));
	goto cleanup;
    }
    if (fcntl_rc>PIPE_SIZE) {
	lcmaps_log(LOG_DEBUG,
	    "%s: current pipe-size (=%d) is larger than preferred size (=%d), "
	    "leaving unchanged\n",
	    __func__, fcntl_rc, PIPE_SIZE);
    } else {
	if ( fcntl(pfd[1], F_SETPIPE_SZ, PIPE_SIZE) == -1 ) {
	    lcmaps_log(LOG_ERR,
		"%s: Error - cannot increase pipesize: (errno=%d, %s)\n",
		__func__, errno, strerror(errno));
	    goto cleanup;
	}
	lcmaps_log(LOG_DEBUG, "%s: pipe-size increased from %d to %d bytes\n",
		__func__, fcntl_rc, PIPE_SIZE);
    }
#endif

    /* Open stream belonging to write-pipe, stored in global static */
    if ( (f_err=fdopen(pfd[1], "w"))==NULL )  {
	lcmaps_log(LOG_ERR,
	    "%s: Error - cannot get stream for pipe: (errno=%d, %s)\n",
	    __func__, errno, strerror(errno));
        goto cleanup;
    }
    
#if PEP_API == 1
    /* Set the log callback first: setting the options will already logged */
    pep_rc=pep_setoption(PEP_OPTION_LOG_HANDLER, pepapi_log);
    if (pep_rc != PEP_OK) {
	lcmaps_log (LOG_ERR,
		    "%s: Error: pep_setoption(PEP_OPTION_LOG_HANDLER, pepapi_log) failed: %d: %s\n",
		    __func__, (int)pep_rc, pep_strerror(pep_rc));
	goto cleanup;
    }
    
    /* PEP-API Log level */
    pep_rc=pep_setoption(PEP_OPTION_LOG_LEVEL, pep_c_loglevel);
    if (pep_rc != PEP_OK) {
	lcmaps_log (LOG_ERR,
		    "%s: Error: pep_setoption(PEP_OPTION_LOG_LEVEL, %d) failed: %d: %s\n",
		    __func__, pep_c_loglevel, (int)pep_rc,pep_strerror(pep_rc));
	goto cleanup;
    }

    /* Set curl to use f_err as its output stream, we have to do this by
     * specifying a log out to the pep library since we don't have access to
     * either cURL or the PEP handle */
    pep_rc=pep_setoption(PEP_OPTION_LOG_STDERR, f_err);
    if (pep_rc != PEP_OK) {
	lcmaps_log (LOG_ERR,
		    "%s: Error: pep_setoption(PEP_OPTION_LOG_STDERR, FILE*) failed: %d: %s\n",
		    __func__, (int)pep_rc, pep_strerror(pep_rc));
	goto cleanup;
    }

#else /* PEP_API */
    /* Set the log callback first: setting the arguments is already logged */
    pep_rc=pep_setoption(pep_handle, PEP_OPTION_LOG_HANDLER, pepapi_log);
    if (pep_rc != PEP_OK) {
	lcmaps_log (LOG_ERR,
		    "%s: Error: pep_setoption(pep_handle, PEP_OPTION_LOG_HANDLER, pepapi_log) failed: %d: %s\n",
		    __func__, (int)pep_rc, pep_strerror(pep_rc));
	goto cleanup;
    }
    
    /* PEP-API Log level */
    pep_rc=pep_setoption(pep_handle, PEP_OPTION_LOG_LEVEL, pep_c_loglevel);
    if (pep_rc != PEP_OK) {
	lcmaps_log (LOG_ERR,
		    "%s: Error: pep_setoption(pep_handle, PEP_OPTION_LOG_LEVEL, %d) failed: %d: %s\n",
		    __func__, pep_c_loglevel, (int)pep_rc,pep_strerror(pep_rc));
	goto cleanup;
    }

    /* Set curl to use f_err as its output stream, we have to do this by
     * specifying a log out to the pep library since we don't have access to
     * either cURL or the PEP handle */
    pep_rc=pep_setoption(pep_handle, PEP_OPTION_LOG_STDERR, f_err);
    if (pep_rc != PEP_OK) {
	lcmaps_log (LOG_ERR,
		    "%s: Error: pep_setoption(pep_handle, PEP_OPTION_LOG_STDERR, FILE*) failed: %d: %s\n",
		    __func__, (int)pep_rc, pep_strerror(pep_rc));
	goto cleanup;
    }
#endif

    /* All is fine, return */
    return 0;

cleanup:
    saved_errno=errno;
    /* Note: pep_setoption(..., PEP_OPTION_LOG_STDERR, f_err); hasn't succeeded,
     * so we can still safely close it */
    pep_close_log_pipe(1);
    errno=saved_errno;

    return -1;
}


/**
 * Logs output from cURL, which is stored in pipe, to LCMAPS log on given level.
 * Returns 0 on success, -1 on error
 */
int pep_log_pipe(int level) {
    static char pipe_buf[PIPE_BUFFER_SIZE];
    static char log_buf[LOG_BUFFER_SIZE];
    char *buf,*pos;
    size_t i,j;
    ssize_t ret;

    /* Check we can actually use the pipe */
    if (pfd[0]==-1 || f_err==NULL)  /* pipe is not open */
	return -1;

    /* First flush the file pointer associated with the output side */
    if (fflush(f_err))	{
	lcmaps_log(LOG_WARNING,"%s: Error calling fflush(): %s\n",
		    __func__, strerror(errno));
    }

    while ((ret = read(pfd[0], pipe_buf, PIPE_BUFFER_SIZE-1)) != 0)
    {
        if (ret == -1)
        {
	    if (errno == EINTR)	/* interrupted, try again */
		continue;
	    else if (errno == EAGAIN) /* nothing there, quit */
		break;
	    else { /* error */
		lcmaps_log(LOG_WARNING,
		    "%s: Error reading from error buffer: %s\n",
		    __func__, strerror(errno));
		return -1;
	    }
        }

        pipe_buf[ret] = '\0';

	/* Break up buffer on newlines, log each line separately */
	buf=pipe_buf;

	while (buf[0]!='\0')  {
	    /* Look for next newline */
	    if ( (pos=strchr(buf,'\n')) )
		pos[0]='\0';

	    /* Replace tabs in buffer by 8 spaces and remove carriage return:
	     * note that HTTP output standard uses carriage-return+linefeed */
	    i=0; j=0;
	    do {
		if (buf[i]=='\t')   { /* Expand tab character */
		    i++;
		    strncpy(&(log_buf[j]),"        ",LOG_BUFFER_SIZE-j);
		    j+=8;
		    if (j>=LOG_BUFFER_SIZE) {
			j=LOG_BUFFER_SIZE-1;
			break;
		    }
		} else if (buf[i]=='\r')    { /* Skip carriage return */
		    i++;
		} else /* Copy other character */
		    log_buf[j++]=buf[i++];
	    } while(buf[i]);
	    /* Properly set \0 char */
	    log_buf[j]='\0';

	    /* Log line */
	    lcmaps_log(level, "%s\n", log_buf);

	    /* when a newline was found, continue */
	    if (pos)
		buf=&(pos[1]);
	    else
		break; /* from while (buf[0]!='\0') */
	}
    }

    return 0;
}


/*
 * Cleanup the pipe and associated file pointer
 * parameter closewrite: when set, close also the write end of the pipe,
 *                       otherwise only close the read end. It will still reset
 *                       the fd and stream to -1 and NULL to allow them to be
 *                       re-used. This hack is necessary since curl will close
 *                       the open filedescriptor, but we won't know it.
 * Returns 0 on success or -1 on error.
 */
void pep_close_log_pipe(int closewrite) {
    int rc;

    /* Now cleanup the file descriptors */

    /* Close read end */
    if (pfd[0] >= 0)	{
	if (close(pfd[0]))
	    lcmaps_log(LOG_WARNING,
		    "%s: warning: closing read side pipe failed: %s\n",
		    __func__, strerror(errno));
	pfd[0]=-1;
    }

    /* Close write end */
    if (pfd[1] >= 0)	{
	if (closewrite)	{
	    if (f_err)  {  /* Use associated stream */
		rc=fclose(f_err);
		f_err=NULL;
		pfd[1]=-1;
	    }
	    else	{	    /* Use underlying file descriptor */
		rc=close(pfd[1]);
		pfd[1]=-1;
	    }
	    if (rc)
		lcmaps_log(LOG_WARNING,
			"%s: warning: closing write side pipe failed: %s\n",
			__func__, strerror(errno));
	} else	{
	    f_err=NULL;
	    pfd[1]=-1;
	}
    }
}


/******************************************************************************
                            Static function definitions
******************************************************************************/
/**
 * pep_log_handler_callback function, catches the PEP library logging and
 * forwards into the LCMAPS logging.
 * Parameters:
 *	level - PEP_LOGLEVEL_ERROR, PEP_LOGLEVEL_WARN, PEP_LOGLEVEL_INFO or
 *	        PEP_LOGLEVEL_DEBUG
 *	fmt   - format string
 *	args  - log arguments
 */
static int pepapi_log(int level, const char *fmt, va_list args)	{
    int len,lvl;
    char *buf;
    va_list args_copy;

    /* Make sure to first flush anything that might be in the pipe from e.g.
     * cURL */
    pep_log_pipe(LOG_DEBUG);

    /* Determine length */
    /* Note: need to use a copy, since vsnprintf() changes the va_list */
    va_copy(args_copy,args); 
    len=vsnprintf(NULL, 0, fmt, args_copy);
    if (len<0)	{
	lcmaps_log(LOG_ERR,
		"%s: Cannot log pep-c log message with fmt \"%s\": %s\n",
		__func__, fmt, strerror(errno));
	va_end(args_copy);
	return -1;
    }
    /* Done with the copy */
    va_end(args_copy);

    /* Allocate memory */
    buf=(char*)malloc((size_t)(len+1));
    if (buf==NULL)  {
	lcmaps_log(LOG_ERR,
		"%s: Out of memory\n", __func__);
	return PEP_ERR_MEMORY;
    }

    /* Make new copy: prefer not to use args itself */
    va_copy(args_copy,args);

    /* Write input */
    if (vsnprintf(buf,(size_t)(len+1),fmt,args_copy)<0)    {
	lcmaps_log(LOG_ERR,
		"%s: Cannot log pep-c log message with fmt \"%s\": %s\n",
		__func__, fmt, strerror(errno));
	free(buf);
	return -1;
    }
    /* Done with the copy */
    va_end(args_copy);

    /* Determine loglevel: PEP_LOGLEVEL_ERROR...PEP_LOGLEVEL_DEBUG */
    switch (level)  {
	case PEP_LOGLEVEL_ERROR:
	    lvl=LOG_ERR; break;
	case PEP_LOGLEVEL_WARN:
	    lvl=LOG_WARNING; break;
	/* Note: PEP library has no NOTICE level */
	case PEP_LOGLEVEL_INFO:
	    lvl=LOG_INFO; break;
	case PEP_LOGLEVEL_DEBUG:
	    lvl=LOG_DEBUG; break;
	default:
	    lcmaps_log(LOG_ERR,"%s: Loglevel %d is unknown\n",__func__,level);
	    free(buf);
	    return -1;
    }
    /* Log the message */
    lcmaps_log(lvl,"%s: %s\n",__func__, buf);
    free(buf);

    return PEP_OK;
}
