/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

/* Define both _XOPEN_SOURCE and _GNU_SOURCE, this makes Solaris happier */
#define _XOPEN_SOURCE 500

/* include lcmaps_c_pep_config.h for PEP_API value, and
 * HAVE_DECL_XACML_RESULT_REMOVEOBLIGATION */
#include "lcmaps_c_pep_config.h"

#if PEP_API == 2 /* Only needed for newer API */
/* Try to provide RTLD_DEFAULT */
/* MacOS needs no POSIX or _DARWIN_C_SOURCE to have RTLD_DEFAULT */
# ifdef __APPLE__
#  define _DARWIN_C_SOURCE
# else
#  define _GNU_SOURCE
# endif
# include <dlfcn.h>
# if !defined(RTLD_DEFAULT) && HAVE_DECL_XACML_RESULT_REMOVEOBLIGATION==1
/* Need to provide xacml_result_removeobligation symbol */
#  include <argus/xacml.h>
# endif
#endif

#include <string.h>
#include <errno.h>

#include <lcmaps/lcmaps_account.h>
#include <lcmaps/lcmaps_utils.h>

#include "pep-c-obligation-handlers.h"
#include "pep-c-obligation-handlers_helpers.h"

#if PEP_API == 1
#define MY_PEP_ERR_OH_PROCESS PEP_ERR_AUTHZ_OH_PROCESS
#else
#define MY_PEP_ERR_OH_PROCESS PEP_ERR_OH_PROCESS
#endif


/* Type of xacml_result_removeobligation() function */
typedef int (*XacmlResultRemoveObligation_t)(xacml_result_t *, int);


/* static pointer to the xacml_result_removeobligation() function, if it exists,
 * or NULL otherwise */
static XacmlResultRemoveObligation_t XacmlResultRemoveObligation = NULL;

/* whether we should ignore unhandled obligations in a Permit result */
static int ignore_unhandled_obligations=0;

/* Sets ignore_unhandled_obligations value */
void set_ignore_unhandled_obligations(int value)    {
    ignore_unhandled_obligations=value;
}

/* Checks for unhandled obligations in the response,
 * returns 0 on success, otherwise -1 */
int unhandled_obligations(xacml_response_t * pepc_response)	{
    xacml_result_t * result;
    xacml_obligation_t * obligation;
    xacml_attributeassignment_t * attribute;
    size_t num_obligations,i,num_attributes;
    int j;
    const char *obligation_id, *attribute_id, *attribute_value;
    int rc=0;
    const char *lcmaps_log_level;
    static int debug=-1;

    /* Have we determined debug mode? */
    if (debug==-1)  {
	lcmaps_log_level=getenv("LCMAPS_DEBUG_LEVEL");
	debug=(lcmaps_log_level && strcmp(lcmaps_log_level,"5")==0 ? 1 : 0);
    }

    /* When ignore_unhandled_obligations is set, we can return straight away */
    if (ignore_unhandled_obligations)
	return rc;

    /* No support for removing obligations: nothing to check */
    if (XacmlResultRemoveObligation==NULL)  {
	lcmaps_log(LOG_WARNING,
		"%s: PEP-API library does not provide "
		"xacml_result_removeobligation(), cannot check whether all "
		"obligations have been fulfilled.\n", __func__);
	return rc;
    }

    result = xacml_response_getresult(pepc_response,0);
    num_obligations=xacml_result_obligations_length(result);
    if (num_obligations!=0) {
	/* Failed to enforce all obligations */
	rc=-1;
	lcmaps_log (LOG_ERR,
		"%s: Result is Permit but %lu unhandled obligation%s "
		"left in the result:\n",
		__func__, (long unsigned)num_obligations,
		(num_obligations==1 ? " is" : "s are"));
	for (i=0; i<num_obligations; i++)   {
	    obligation=xacml_result_getobligation(result,(int)i);
	    obligation_id=xacml_obligation_getid(obligation);
	    lcmaps_log(LOG_ERR, "%s:   %s\n",
		__func__, obligation_id ? obligation_id : "(NULL)");
	    if (debug)	{ /* Also log attributes */
		num_attributes=xacml_obligation_attributeassignments_length(obligation);
		for (j=0; (size_t)j<num_attributes; j++)    {
		    attribute=xacml_obligation_getattributeassignment(obligation, j);
		    attribute_id=xacml_attributeassignment_getid(attribute);
		    attribute_value=xacml_attributeassignment_getvalue(attribute);
		    lcmaps_log(LOG_DEBUG,
			    "%s:    attr_assignment %d: %s=%s\n",
			    __func__, j, 
			    (attribute_id ? attribute_id : "(NULL)"),
			    (attribute_value ? attribute_value : "(NULL)"));
		}
	    }
	}
    }

    return rc;
}

/* Checks consistency of the resulting uid/pgids. When ok, stores them into the
 * LCMAPS framework.
 * returns 0 on success, otherwise -1 */
int parse_and_store_creddata(void)   {
    uid_t *uid=NULL;
    gid_t *pgid=NULL;
    gid_t *sgid=NULL;
    size_t i, uid_cnt=0, pgid_cnt=0, sgid_cnt=0;

    /* Get resulting data */
    get_uid_list(&uid, &uid_cnt);
    get_pgid_list(&pgid, &pgid_cnt);
    get_sgid_list(&sgid, &sgid_cnt);

    /* Check consistency UID(s), we allow 0 UID(s) */
    if (uid_cnt>1)  { /* More than 1 uid: need to check consistency */
	for (i=1; i<uid_cnt; i++)	{
	    if (uid[i]!=uid[0]) {
		lcmaps_log(LOG_ERR,
		    "%s: XACML: Inconsistent mapping result: "
		    "multiple and different UIDs returned.\n",
		    __func__);
		return -1;
	    }
	}
    }

    /* Check consistency primary GID(s), we allow 0 UID(s) */
    if (pgid_cnt>1)  { /* More than 1 pGID: need to check consistency */
	for (i=1; i<pgid_cnt; i++)	{
	    if (pgid[i]!=pgid[0]) {
		lcmaps_log(LOG_ERR,
		    "%s: XACML: Inconsistent mapping result: "
		    "multiple and different pGIDs returned.\n",
		    __func__);
		return -1;
	    }
	}
    }

    /* Add (first) UID */
    if (uid_cnt && addCredentialData(UID, uid))    {
	lcmaps_log(LOG_ERR, "%s: Error adding UID %d\n", __func__, uid[0]);
	return -1;
    }

    /* Add (first) pGID */
    if (pgid_cnt && addCredentialData(PRI_GID, pgid))    {
	lcmaps_log(LOG_ERR, "%s: Error adding pGID %d\n", __func__, pgid[0]);
	return -1;
    }

    /* secondary GID(s): no consistency checks, just add them all */
    for (i=0; i<sgid_cnt; i++)	{
	if (addCredentialData(SEC_GID, &(sgid[i]))) {
	    lcmaps_log(LOG_ERR,"%s: Error adding sGID %d\n", __func__, sgid[i]);
	    return -1;
	}
    }

    /* All done */
    return 0;
}

/* registers the obligation handlers, result is returned in parameters, return
 * value is 0 on success or -1 on error */
int registerObligationHandlers (pep_obligationhandler_t ** oh, size_t * oh_count)
{
    pep_obligationhandler_t *oh_tmp = NULL;
    const size_t oh_max=5;

    /* Initialize oh and oh_count */
    *oh=NULL; *oh_count=0;

    /* Allocate the array for pointers to obligation handlers */
    if ( (oh_tmp = (pep_obligationhandler_t *)calloc (oh_max, sizeof (pep_obligationhandler_t))) == NULL)
	return -1;

    /* Most UNIX now support RTLD_DEFAULT (POSIX reserved) */
#ifdef RTLD_DEFAULT
    /* Clear dlerror() */
    dlerror();
    /* Try to load the function */
    XacmlResultRemoveObligation=
	(XacmlResultRemoveObligation_t)dlsym(RTLD_DEFAULT,
					     "xacml_result_removeobligation");
    if (XacmlResultRemoveObligation==NULL)  {
	/* Do not warn here, we do that at the end in case of a permit */
	lcmaps_log(LOG_DEBUG,
		"%s: PEP-C library does not provide xacml_result_removeobligation() symbol: %s.\n",
		__func__, dlerror());
    }
#elif HAVE_DECL_XACML_RESULT_REMOVEOBLIGATION == 1
    /* No RTLD_DEFAULT but symbol existed at least at build-time */
    lcmaps_log(LOG_DEBUG,
	    "%s: platform does not understand RTLD_DEFAULT, build-time PEP-C library supports xacml_result_removeobligation().\n",
	    __func__);
    XacmlResultRemoveObligation=xacml_result_removeobligation;
#else
    /* No RTLD_DEFAULT and no xacml_result_removeobligation at build-time */
    lcmaps_log(LOG_DEBUG,
	    "%s: plugin is built against PEP-C library without support for xacml_result_removeobligation.\n");
#endif

    /* Setup and link to the local-environment-map/posix Obligation */
    oh_tmp[0].id      = strdup(XACML_GRIDWN_OBLIGATION_LOCAL_ENVIRONMENT_MAP_POSIX);
    oh_tmp[0].init    = oh_init_local_environment_map_posix;
    oh_tmp[0].process = oh_process_local_environment_map_posix;
    oh_tmp[0].destroy = oh_destroy_local_environment_map_posix;

    add_supported_obligation_to_list (XACML_GRIDWN_OBLIGATION_LOCAL_ENVIRONMENT_MAP_POSIX);

    /* Setup and link to the UID GID Obligation */
    oh_tmp[1].id      = strdup(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID);
    oh_tmp[1].init    = oh_init_uidgid;
    oh_tmp[1].process = oh_process_uidgid;
    oh_tmp[1].destroy = oh_destroy_uidgid;

    add_supported_obligation_to_list (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID);

    
    /* Setup and link to the Secondary GID Obligation */
    oh_tmp[2].id      = strdup(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_SecondaryGIDs);
    oh_tmp[2].init    = oh_init_secondarygids;
    oh_tmp[2].process = oh_process_secondarygids;
    oh_tmp[2].destroy = oh_destroy_secondarygids;

    add_supported_obligation_to_list (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_SecondaryGIDs);


    /* Setup and link to the Username Obligation */
    oh_tmp[3].id      = strdup(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Username);
    oh_tmp[3].init    = oh_init_username;
    oh_tmp[3].process = oh_process_username;
    oh_tmp[3].destroy = oh_destroy_username;

    add_supported_obligation_to_list (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Username);

    /* Setup and link to the Account Obligation */
    oh_tmp[4].id      = strdup(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Account);
    oh_tmp[4].init    = oh_init_account;
    oh_tmp[4].process = oh_process_account;
    oh_tmp[4].destroy = oh_destroy_account;

    add_supported_obligation_to_list (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Account);

    *oh = oh_tmp;
    *oh_count=oh_max;
    
    /* Print the registered supported obligations */
    print_supported_obligations_list();
    lcmaps_log(LOG_DEBUG, "registerObligationHandlers finished\n");
    
    return 0;
}



int oh_init_uidgid (void)
{
    lcmaps_log(LOG_DEBUG, "oh_init_uidgid()\n");
    return PEP_OK;
}

int oh_init_secondarygids (void)
{
    lcmaps_log(LOG_DEBUG, "oh_init_secondarygids()\n");
    return PEP_OK;
}

int oh_init_username (void)
{
    lcmaps_log(LOG_DEBUG, "oh_init_username()\n");
    return PEP_OK;
}

int oh_init_account (void)
{
    lcmaps_log(LOG_DEBUG, "oh_init_account()\n");
    return PEP_OK;
}

int oh_init_local_environment_map_posix (void)
{
    lcmaps_log(LOG_DEBUG, "oh_init_local_environment_map_posix()\n");
    return PEP_OK;
}


int oh_process_uidgid (xacml_request_t ** this_request, xacml_response_t ** this_response)
{
    size_t i, j, k, results_l, obligations_l, attrs_l;
    int curr_oblig;
    long var_long = -1;
    xacml_response_t * response  = *this_response;
    xacml_result_t * result;
    xacml_obligation_t * obligation;
    xacml_attributeassignment_t * attr;
    xacml_fulfillon_t fulfillon;
    const char * obligation_id  = NULL;
    const char * attr_id  = NULL;
    const char *attr_value = NULL;
    char *endptr = NULL;
    size_t count_attrib_uid   = 0;
    size_t count_attrib_p_gid = 0;

    lcmaps_log(LOG_DEBUG, "%s\n", __func__);

    if (response == NULL) {
        lcmaps_log(LOG_DEBUG, "%s: response is NULL\n", __func__);
        return PEP_ERR_NULL_POINTER;
    }

    /* Check response message and verify that the amount of results and the
     * content of the results is OK */
    if (C_PEP_OK != checkResponseSanity (response)) {
	/* Don't log this on LOG_ERR: it can also be because of a Deny, errors
	 * are already logged by checkResponseSanity() itself too */
        lcmaps_log (LOG_INFO,
		"%s: checkResponseSanity() returned a failure condition in the response message. "
		"Stopped looking into the obligations\n", __func__);
        return MY_PEP_ERR_OH_PROCESS;
    } else {
        lcmaps_log(LOG_DEBUG,
		"%s: checkResponseSanity() returned OK. "
		"Continuing with the search for obligation information\n",
		__func__);
    }
        
    /* Going to obligations blob, skip the rest (as it was tested in detail in
     * checkResponseSanity()) */
    results_l = xacml_response_results_length(response);
    for (i = 0; i < results_l; i++) {
        result = xacml_response_getresult(response,(int)i);
        
        obligations_l = xacml_result_obligations_length(result);
        lcmaps_log(LOG_DEBUG, "%s: response.result[%lu]: %lu obligations\n",
		   __func__, (long unsigned)i, (long unsigned)obligations_l);
        
        /* Hunt for the obligation id of interest, skip the rest */
        for (j=0, curr_oblig=0; j < obligations_l; j++) {
	    /* Get current obligation: cannot use j, since we might remove
	     * obligations, resulting in a counter mismatch. */
            obligation = xacml_result_getobligation(result,curr_oblig);
	    obligation_id = xacml_obligation_getid(obligation);

            /*
             * Searching for the Obligation ID: 
             * XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID 
             */
            if (strcmp(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID, obligation_id) != 0)
            {
                lcmaps_log(LOG_DEBUG,
			"%s: response.result[%lu].obligation[%lu].id = \"%s\". "
			"Skipping non-applicable Obligation ID\n",
			__func__, (long unsigned)i, (long unsigned)j,
			obligation_id);
		curr_oblig++;
                continue;
            }

	    /* Check that we should run on Permit */
	    fulfillon=xacml_obligation_getfulfillon(obligation);
	    if (fulfillon!=XACML_FULFILLON_PERMIT)  {
		lcmaps_log(LOG_DEBUG,
			"%s:  response.result[%lu].obligation[%lu].id = \"%s\". "
			"Skipping obligation that should run on %s\n",
			__func__, (long unsigned)i, (long unsigned)j,
			obligation_id, fulfillon_str(fulfillon));
		curr_oblig++;
		continue;
	    }

            /* We've gottum! */
            lcmaps_log(LOG_DEBUG,
		    "%s: response.result[%lu].obligation[%lu].id = \"%s\". "
		    "Found applicable Obligation ID\n",
		    __func__, (long unsigned)i, (long unsigned)j, obligation_id);

            attrs_l = xacml_obligation_attributeassignments_length(obligation);
            lcmaps_log(LOG_DEBUG,
		    "%s: response.result[%lu].obligation[%lu]: %lu attributes.\n",
		    __func__, (long unsigned)i, (long unsigned)j,(long unsigned)attrs_l);

	    /* Reset attribute counters */
	    count_attrib_uid = count_attrib_p_gid = 0;

            for (k= 0; k < attrs_l; k++) {
                /* Reading individual attributes within obligations */
                attr = xacml_obligation_getattributeassignment(obligation, (int)k);
		attr_id = xacml_attributeassignment_getid(attr);

                /* Is the current value a POSIX Unix UID */
                if (strcmp (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID, attr_id) == 0)
                {
                    /* Check if we didn't already had this attribute id... */
                    if (count_attrib_uid == MAX_APPEARANCE_OF_ATTRIB_UID) {
                        lcmaps_log (LOG_ERR,
				"%s: Error: more than %d attribute of type \"%s\".\n", 
                                __func__, MAX_APPEARANCE_OF_ATTRIB_UID,
				XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID);
                        return MY_PEP_ERR_OH_PROCESS;
                    }

		    attr_value = xacml_attributeassignment_getvalue(attr);
		    if (attr_value==NULL || *attr_value=='\0') {
			lcmaps_log (LOG_ERR,
			    "%s: Error: empty value at response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__,  (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
			errno = EINVAL;
			return MY_PEP_ERR_OH_PROCESS;
		    }
		   
		    /* Log id/value */
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].value= %s\n", 
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_value);
		    
		    /* Convert string to long */
		    errno = 0;
		    var_long = strtol (attr_value, &endptr, 10);
		    if (*endptr!='\0' || var_long<0)
			errno=EINVAL;
		    if (errno)	{
			lcmaps_log (LOG_ERR,
				"%s: Error: couldn't parse the value %s at response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n", 
				__func__, attr_value, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
			return MY_PEP_ERR_OH_PROCESS;
		    }
		    
		    /* Add to internal list */
		    if (addUid((uid_t) var_long))
			return MY_PEP_ERR_OH_PROCESS;

		    /* Administrative, counting appearance of attribute-id
		     * XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID */
		    count_attrib_uid++;
                }
                /* Is the current value a POSIX Unix GID */
                else if (strcmp (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID, attr_id) == 0)
                {
                    /* Check if we didn't already had this attribute id... */
                    if (count_attrib_p_gid == MAX_APPEARANCE_OF_ATTRIB_PRI_GID) {
                        lcmaps_log (LOG_ERR,
				"%s: Error: more than %d attribute of type \"%s\".\n", 
                                __func__, MAX_APPEARANCE_OF_ATTRIB_PRI_GID,
				XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID);
                        return MY_PEP_ERR_OH_PROCESS;
                    }

		    attr_value = xacml_attributeassignment_getvalue(attr);
		    if (attr_value==NULL || *attr_value=='\0') {
			lcmaps_log (LOG_ERR,
			    "%s: Error: empty value at response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__,  (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
			errno = EINVAL;
			return MY_PEP_ERR_OH_PROCESS;
		    }
		    
		    /* Log id/value */
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].value= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_value);

		    /* Convert string to long */
		    errno = 0;
		    var_long = strtol (attr_value, &endptr, 10);
		    if (*endptr!='\0' || var_long<0)
			errno=EINVAL;
		    if (errno) {
			lcmaps_log (LOG_ERR,
				"%s: Error: couldn't parse the value %s at response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n", 
				__func__, attr_value, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
			return MY_PEP_ERR_OH_PROCESS;
		    }

		    /* Add to internal list */
		    if (addPGid((gid_t)var_long)<0)
			return MY_PEP_ERR_OH_PROCESS;
                        
		    /* Administrative, counting appearance of attribute-id
		     * XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID */
                    count_attrib_p_gid++;
                }
		/* Unknown attribute */
                else
                {
                    lcmaps_log (LOG_ERR,
			    "%s: Error: Unexpected attribute: %s\n",
                            __func__, attr_id);
                    return MY_PEP_ERR_OH_PROCESS;
                }
            }

	    /* Now remove the handled obligation or update curr_oblig */
	    if (XacmlResultRemoveObligation)
		XacmlResultRemoveObligation(result,curr_oblig);
	    else
		curr_oblig++;
        }
    }

    return PEP_OK;
}

int oh_process_secondarygids (xacml_request_t ** this_request, xacml_response_t ** this_response)
{
    size_t i, j, k, results_l, obligations_l, attrs_l;
    int curr_oblig;
    xacml_response_t * response  = *this_response;
    xacml_result_t * result;
    xacml_obligation_t * obligation;
    xacml_attributeassignment_t * attr;
    xacml_fulfillon_t fulfillon;
    const char * obligation_id  = NULL;
    const char * attr_id  = NULL;
    const char *attr_value = NULL;
    char *endptr = NULL;
    long var_long;

    lcmaps_log(LOG_DEBUG, "%s\n", __func__);

    if (response == NULL) {
        lcmaps_log(LOG_DEBUG, "%s: response is NULL\n", __func__);
        return PEP_ERR_NULL_POINTER;
    }

    /* Check response message and verify that the amount of results and the
     * content of the results is OK */
    if (C_PEP_OK != checkResponseSanity (response)) {
	/* Don't log this on LOG_ERR: it can also be because of a Deny, errors
	 * are already logged by checkResponseSanity() itself too */
        lcmaps_log (LOG_INFO,
		"%s: checkResponseSanity() returned a failure condition in the response message. "
		"Stopped looking into the obligations\n", __func__);
        return MY_PEP_ERR_OH_PROCESS;
    } else {
        lcmaps_log(LOG_DEBUG,
		"%s: checkResponseSanity() returned OK. "
		"Continuing with the search for obligation information\n",
		__func__);
    }
        
    /* Going to obligations blob, skip the rest (as it was tested in detail in
     * checkResponseSanity()) */
    results_l = xacml_response_results_length(response);
    for (i = 0; i < results_l; i++) {
        result = xacml_response_getresult(response,(int)i);
        
        obligations_l = xacml_result_obligations_length(result);
        lcmaps_log(LOG_DEBUG, "%s: response.result[%lu]: %lu obligations\n",
		   __func__, (long unsigned)i, (long unsigned)obligations_l);
        
        /* Hunt for the obligation id of interest, skip the rest */
        for (j=0, curr_oblig=0; j < obligations_l; j++) {
	    /* Get current obligation: cannot use j, since we might remove
	     * obligations, resulting in a counter mismatch. */
            obligation = xacml_result_getobligation(result,curr_oblig);
	    obligation_id = xacml_obligation_getid(obligation);

            /*
             * Searching for the Obligation ID: 
             * XACML_ATTR_PROFILE_NS_URL_OBLIGATION_SecondaryGIDs
             */
            if (strcmp(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_SecondaryGIDs, obligation_id) != 0)
            {
                lcmaps_log(LOG_DEBUG,
			"%s: response.result[%lu].obligation[%lu].id = \"%s\". "
			"Skipping non-applicable Obligation ID\n",
                        __func__, (long unsigned)i, (long unsigned)j,
			obligation_id);
		curr_oblig++;
                continue;
            }

	    /* Check that we should run on Permit */
	    fulfillon=xacml_obligation_getfulfillon(obligation);
	    if (fulfillon!=XACML_FULFILLON_PERMIT)  {
		lcmaps_log(LOG_DEBUG,
			"%s:  response.result[%lu].obligation[%lu].id = \"%s\". "
			"Skipping obligation that should run on %s\n",
			__func__, (long unsigned)i, (long unsigned)j,
			obligation_id, fulfillon_str(fulfillon));
		curr_oblig++;
		continue;
	    }

            /* We've gottum! */
            lcmaps_log(LOG_DEBUG,
		    "%s: response.result[%lu].obligation[%lu].id = \"%s\". "
		    "Found applicable Obligation ID\n",
		    __func__, (long unsigned)i, (long unsigned)j, obligation_id);

            attrs_l = xacml_obligation_attributeassignments_length(obligation);
            lcmaps_log(LOG_DEBUG,
		    "%s: response.result[%lu].obligation[%lu]: %lu attributes.\n",
		    __func__, (long unsigned)i, (long unsigned)j,(long unsigned)attrs_l);

            for (k= 0; k < attrs_l; k++) {
                /* Reading individual attributes within obligations */
                attr = xacml_obligation_getattributeassignment(obligation, (int)k);
		attr_id = xacml_attributeassignment_getid(attr);

                /* Is the current value a POSIX Unix GID */
                if (strcmp (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID, attr_id) == 0)
                {
		    attr_value = xacml_attributeassignment_getvalue(attr);
		    if (attr_value==NULL || *attr_value=='\0') {
			lcmaps_log (LOG_ERR,
			    "%s: Error: empty value at response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__,  (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
			errno = EINVAL;
			return MY_PEP_ERR_OH_PROCESS;
		    }
		    
		    /* Log id/value */
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].value= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_value);

		    /* Convert string to long */
		    errno = 0;
		    var_long = strtol (attr_value, &endptr, 10);
		    if (*endptr!='\0' || var_long<0)
			errno=EINVAL;
		    if (errno) {
			lcmaps_log (LOG_ERR,
				"%s: Error: couldn't parse the value %s at response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n", 
				__func__, attr_value, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
			return MY_PEP_ERR_OH_PROCESS;
		    }
		    
		    /* Add to internal list */
		    if (addSGid((gid_t*)&var_long,1)<0)
			return MY_PEP_ERR_OH_PROCESS;
                }
		/* Unknown attribute */
                else
                {
                    lcmaps_log (LOG_ERR,
			    "%s: Error: Unexpected attribute: %s\n",
			    __func__, attr_id);
                    return MY_PEP_ERR_OH_PROCESS;
                }
            }

	    /* Now remove the handled obligation or update curr_oblig */
	    if (XacmlResultRemoveObligation)
		XacmlResultRemoveObligation(result,curr_oblig);
	    else
		curr_oblig++;
        }
    }

    return PEP_OK;
}

int oh_process_username (xacml_request_t ** this_request, xacml_response_t ** this_response)
{
    size_t i, j, k, results_l, obligations_l, attrs_l;
    int curr_oblig;
    xacml_response_t * response  = *this_response;
    xacml_result_t * result;
    xacml_obligation_t * obligation;
    xacml_attributeassignment_t * attr;
    xacml_fulfillon_t fulfillon;
    const char * obligation_id  = NULL;
    const char * attr_id  = NULL;
    size_t count_attrib_username  = 0;
    const char * username = NULL;
    struct passwd * pw = NULL;
    int cnt_sec_gid = 0;
    gid_t * sec_gid = NULL;

    lcmaps_log(LOG_DEBUG, "%s\n", __func__);

    if (response == NULL) {
        lcmaps_log(LOG_DEBUG, "%s: response is NULL\n", __func__);
        return PEP_ERR_NULL_POINTER;
    }

    /* Check response message and verify that the amount of results and the
     * content of the results is OK */
    if (C_PEP_OK != checkResponseSanity (response)) {
	/* Don't log this on LOG_ERR: it can also be because of a Deny, errors
	 * are already logged by checkResponseSanity() itself too */
        lcmaps_log (LOG_INFO,
		"%s: checkResponseSanity() returned a failure condition in the response message. "
		"Stopped looking into the obligations\n", __func__);
        return MY_PEP_ERR_OH_PROCESS;
    } else {
        lcmaps_log(LOG_DEBUG,
		"%s: checkResponseSanity() returned OK. "
		"Continuing with the search for obligation information\n",
		__func__);
    }
        
    /* Going to obligations blob, skip the rest (as it was tested in detail in
     * checkResponseSanity()) */
    results_l = xacml_response_results_length(response);
    for (i = 0; i < results_l; i++) {
        result = xacml_response_getresult(response,(int)i);
        
        obligations_l = xacml_result_obligations_length(result);
        lcmaps_log(LOG_DEBUG, "%s: response.result[%lu]: %lu obligations\n",
		   __func__, (long unsigned)i, (long unsigned)obligations_l);
        
        /* Hunt for the obligation id of interest, skip the rest */
        for (j=0, curr_oblig=0; j < obligations_l; j++) {
	    /* Get current obligation: cannot use j, since we might remove
	     * obligations, resulting in a counter mismatch. */
            obligation = xacml_result_getobligation(result,curr_oblig);
	    obligation_id = xacml_obligation_getid(obligation);

            /*
             * Searching for the Obligation ID: 
             * XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Username     
             */
            if (strcmp(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Username, obligation_id) != 0)
            {
                lcmaps_log(LOG_DEBUG,
			"%s: response.result[%lu].obligation[%lu].id = \"%s\". "
			"Skipping non-applicable Obligation ID\n",
			__func__, (long unsigned)i, (long unsigned)j,
			obligation_id);
		curr_oblig++;
                continue;
            }

	    /* Check that we should run on Permit */
	    fulfillon=xacml_obligation_getfulfillon(obligation);
	    if (fulfillon!=XACML_FULFILLON_PERMIT)  {
		lcmaps_log(LOG_DEBUG,
			"%s:  response.result[%lu].obligation[%lu].id = \"%s\". "
			"Skipping obligation that should run on %s\n",
			__func__, (long unsigned)i, (long unsigned)j,
			obligation_id, fulfillon_str(fulfillon));
		curr_oblig++;
		continue;
	    }

            /* We've gottum! */
            lcmaps_log(LOG_DEBUG,
		    "%s: response.result[%lu].obligation[%lu].id = \"%s\". "
		    "Found applicable Obligation ID\n",
		    __func__, (long unsigned)i, (long unsigned)j, obligation_id);

            attrs_l = xacml_obligation_attributeassignments_length(obligation);
            lcmaps_log(LOG_DEBUG,
		    "%s: response.result[%lu].obligation[%lu]: %lu attributes.\n",
		    __func__, (long unsigned)i, (long unsigned)j,(long unsigned)attrs_l);

	    /* Reset attribute counters */
	    count_attrib_username = 0;

            for (k= 0; k < attrs_l; k++) {
                /* Reading individual attributes within obligations */
                attr = xacml_obligation_getattributeassignment(obligation, (int)k);
		attr_id = xacml_attributeassignment_getid(attr);

                /* Is the current value a Username */
                if (strcmp (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_USERNAME, attr_id) == 0)
                {
                    /* Check if we didn't already had this attribute id... */
		    /* The maximum number of Usernames in an obligation is 1,
		     * and only 1 */
                    if (count_attrib_username == 1) {
                        lcmaps_log (LOG_ERR,
				"%s: Error: more than 1 attribute of type \"%s\".\n", 
                                __func__, XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_USERNAME);
                        return MY_PEP_ERR_OH_PROCESS;
                    }

		    /* Got the username, now to convert it into UIDs and GIDs
		     * through the POSIX account system */
		    username = xacml_attributeassignment_getvalue(attr);
		    if (username==NULL || *username=='\0') {
			lcmaps_log (LOG_ERR,
			    "%s: Error: empty value at response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__,  (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
			errno = EINVAL;
			return MY_PEP_ERR_OH_PROCESS;
		    }

		    /* Log id/value */
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].value= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, username);

		    pw = getpwnam(username);
		    if (!pw) {
			lcmaps_log (LOG_ERR,
				"%s: Error: getpwnam() did not find the username '%s'.\n",
				__func__, username);
			return MY_PEP_ERR_OH_PROCESS;
		    }

		    /* Add to internal list */
		    if (addUid(pw -> pw_uid)<0)
			return MY_PEP_ERR_OH_PROCESS;
		    if (addPGid(pw -> pw_gid)<0)
			return MY_PEP_ERR_OH_PROCESS;

		    /*
		     * Retrieve secondary group id's
		     */
		    if (lcmaps_get_gidlist(username, &cnt_sec_gid, &sec_gid)==0)
		    {
			if (addSGid(sec_gid, (size_t)cnt_sec_gid)<0)	{
			    free(sec_gid);
			    return MY_PEP_ERR_OH_PROCESS;
			}
			free(sec_gid);
		    }

		    /* Administrative, counting appearance of attribute-id
		     * XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID */
		    count_attrib_username++;
                }
		/* Unknown attribute */
                else
                {
                    lcmaps_log (LOG_ERR,
			    "%s: Error: Unexpected attribute: %s\n",
			    __func__, attr_id);
                    return MY_PEP_ERR_OH_PROCESS;
                }
            }

	    /* Now remove the handled obligation or update curr_oblig */
	    if (XacmlResultRemoveObligation)
		XacmlResultRemoveObligation(result,curr_oblig);
	    else
		curr_oblig++;
        }
    }

    return PEP_OK;
}

int oh_process_account (xacml_request_t ** this_request, xacml_response_t ** this_response)
{
    size_t i, j, k, results_l, obligations_l, attrs_l;
    int curr_oblig;
    xacml_response_t * response = *this_response;
    xacml_result_t * result;
    xacml_obligation_t * obligation;
    xacml_attributeassignment_t * attr;
    xacml_fulfillon_t fulfillon;
    const char * obligation_id  = NULL;
    const char * attr_id        = NULL;
    size_t count_attrib_username    = 0;
    size_t count_attrib_primgroupid = 0;
    const char * username       = NULL;
    const char * groupid        = NULL;
    const char * secgroupid     = NULL;
    uid_t this_uid;
    gid_t this_prigid;
    gid_t this_secgid;

    lcmaps_log(LOG_DEBUG, "%s\n", __func__);

    if (response == NULL) {
        lcmaps_log(LOG_DEBUG, "%s: response is NULL\n", __func__);
        return PEP_ERR_NULL_POINTER;
    }

    /* Check response message and verify that the amount of results and the
     * content of the results is OK */
    if (C_PEP_OK != checkResponseSanity (response)) {
	/* Don't log this on LOG_ERR: it can also be because of a Deny, errors
	 * are already logged by checkResponseSanity() itself too */
        lcmaps_log (LOG_INFO,
		"%s: checkResponseSanity() returned a failure condition in the response message. "
		"Stopped looking into the obligations\n", __func__);
        return MY_PEP_ERR_OH_PROCESS;
    } else {
        lcmaps_log(LOG_DEBUG,
		"%s: checkResponseSanity() returned OK. "
		"Continuing with the search for obligation information\n",
		__func__);
    }
        
    /* Going to obligations blob, skip the rest (as it was tested in detail in
     * checkResponseSanity()) */
    results_l = xacml_response_results_length(response);
    for (i = 0; i < results_l; i++) {
        result = xacml_response_getresult(response,(int)i);
        
        obligations_l = xacml_result_obligations_length(result);
        lcmaps_log(LOG_DEBUG, "%s: response.result[%lu]: %lu obligations\n",
		   __func__, (long unsigned)i, (long unsigned)obligations_l);
        
        /* Hunt for the obligation id of interest, skip the rest */
        for (j=0, curr_oblig=0; j < obligations_l; j++) {
	    /* Get current obligation: cannot use j, since we might remove
	     * obligations, resulting in a counter mismatch. */
            obligation = xacml_result_getobligation(result,curr_oblig);
	    obligation_id = xacml_obligation_getid(obligation);

            /*
             * Searching for the Obligation ID: 
             * XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Account 
             */
            if (strcmp(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Account, obligation_id) != 0)
            {
                lcmaps_log(LOG_DEBUG,
			"%s: response.result[%lu].obligation[%lu].id = \"%s\". "
			"Skipping non-applicable Obligation ID\n",
			__func__, (long unsigned)i, (long unsigned)j,
			obligation_id);
		curr_oblig++;
                continue;
            }

	    /* Check that we should run on Permit */
	    fulfillon=xacml_obligation_getfulfillon(obligation);
	    if (fulfillon!=XACML_FULFILLON_PERMIT)  {
		lcmaps_log(LOG_DEBUG,
			"%s:  response.result[%lu].obligation[%lu].id = \"%s\". "
			"Skipping obligation that should run on %s\n",
			__func__, (long unsigned)i, (long unsigned)j,
			obligation_id, fulfillon_str(fulfillon));
		curr_oblig++;
		continue;
	    }

            /* We've gottum! */
            lcmaps_log(LOG_DEBUG,
		    "%s: response.result[%lu].obligation[%lu].id = \"%s\". "
		    "Found applicable Obligation ID\n",
		    __func__, (long unsigned)i, (long unsigned)j, obligation_id);

            attrs_l = xacml_obligation_attributeassignments_length(obligation);
            lcmaps_log(LOG_DEBUG,
		    "%s: response.result[%lu].obligation[%lu]: %lu attributes.\n",
		    __func__, (long unsigned)i, (long unsigned)j,(long unsigned)attrs_l);

	    /* Reset attribute counters */
	    count_attrib_username = count_attrib_primgroupid = 0;

            for (k= 0; k < attrs_l; k++) {
                /* Reading individual attributes within obligations */
                attr = xacml_obligation_getattributeassignment(obligation, (int)k);
		attr_id = xacml_attributeassignment_getid(attr);

                /* Is the current value a User-ID */
                if (strcmp (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_USERNAME, attr_id) == 0)
                {
                    /* Check if we didn't already had this attribute id... */
		    /* The maximum number of Usernames in an obligation is 1,
		     * and only 1 */
                    if (count_attrib_username == 1) {
                        lcmaps_log (LOG_ERR,
				"%s: Error: more than 1 attribute of type \"%s\".\n",
                                __func__, attr_id);
                        return MY_PEP_ERR_OH_PROCESS;
                    }

		    /* Got the username, now to convert it into UIDs through the
		     * POSIX account system */
		    username = xacml_attributeassignment_getvalue(attr);
		    if (username==NULL || *username=='\0') {
			lcmaps_log (LOG_ERR,
			    "%s: Error: empty value at response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__,  (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
			errno = EINVAL;
			return MY_PEP_ERR_OH_PROCESS;
		    }
		    
		    /* Log id/value */
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].value= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, username);

		    /* Converting the User-ID into a UID and registering the
		     * values into the LCMAPS framework */
		    if (threadsafe_getuid_from_name (username, &this_uid) < 0) {
			lcmaps_log (LOG_ERR,
				"%s: Error: username not found: \"%s\".\n",
				__func__, username);
			return MY_PEP_ERR_OH_PROCESS;
		    }
		    
		    /* Add to internal list */
		    if (addUid(this_uid)<0)
			return MY_PEP_ERR_OH_PROCESS;

		    /* Administrative, counting appearance of attribute-id
		     * XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_USERNAME
		     * */
                    count_attrib_username++;
                }
                /* Is the current value a Primary Group-ID */
                else if (strcmp (XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_PRIMARY_GROUPNAME, attr_id) == 0)
                {
                    /* Check if we didn't already had this attribute id... */
                    /* The maximum number of Primary Groups in an obligation is 1, and only 1 */
                    if (count_attrib_primgroupid == 1) {
                        lcmaps_log (LOG_ERR,
				"%s: Error: more than 1 attribute \"%s\".\n", 
                                __func__, attr_id);
                        return MY_PEP_ERR_OH_PROCESS;
                    }
		    
		    /* Got the groupname, now to convert it into a GID through
		     * the POSIX account system */
		    groupid = xacml_attributeassignment_getvalue(attr);
		    if (groupid==NULL || *groupid=='\0') {
			lcmaps_log (LOG_ERR,
			    "%s: Error: empty value at response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__,  (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
			errno = EINVAL;
			return MY_PEP_ERR_OH_PROCESS;
		    }

		    /* Log id/value */
		    lcmaps_log (LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].value= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, groupid);

		    /* Converting the Group-ID into a GID */
		    if (threadsafe_getgid_from_name (groupid, &this_prigid)<0)
		    {
			lcmaps_log (LOG_ERR, "%s: Error: couldn't find the group entry for the Group-ID \"%s\".\n", __func__, groupid);
			return MY_PEP_ERR_OH_PROCESS;
		    }
		    
		    /* Add to internal list */
		    if (addPGid(this_prigid)<0)
			return MY_PEP_ERR_OH_PROCESS;
		    
		    /* Administrative, counting appearance of attribute-id
		     * XACML_GRIDWN_ATTRIBUTE_GROUP_ID */
		    count_attrib_primgroupid++;
                }
                /* Is the current value a Secondary Group-ID */
                else if (strcmp(XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_SECONDARY_GROUPNAME, attr_id) == 0)
                {
		    /* Got the groupname, now to convert it into a GID through
		     * the POSIX account system */
		    secgroupid = xacml_attributeassignment_getvalue(attr);
		    if (secgroupid==NULL || *secgroupid=='\0') {
			lcmaps_log (LOG_ERR,
			    "%s: Error: empty value at response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__,  (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
			errno = EINVAL;
			return MY_PEP_ERR_OH_PROCESS;
		    }

		    /* Log id/value */
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].value= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, secgroupid);

		    /* Converting the Group-ID into a GID */
		    if ( threadsafe_getgid_from_name (secgroupid, &this_secgid)<0)
		    {
			lcmaps_log (LOG_ERR,
				"%s: Error: group-id \"%s\".\n", __func__, secgroupid);
			return MY_PEP_ERR_OH_PROCESS;
		    }
		    
		    /* Add to internal list */
		    if (addSGid(&this_secgid,1)<0)
			return MY_PEP_ERR_OH_PROCESS;
                }
		/* Unknown attribute */
                else
                {
                    lcmaps_log (LOG_ERR,
			    "%s: Error: Unexpected attribute: %s\n",
			    __func__, attr_id);
                    return MY_PEP_ERR_OH_PROCESS;
                }
            }

	    /* Now remove the handled obligation or update curr_oblig */
	    if (XacmlResultRemoveObligation)
		XacmlResultRemoveObligation(result,curr_oblig);
	    else
		curr_oblig++;
        }
    }

    return PEP_OK;
}

int oh_process_local_environment_map_posix (xacml_request_t ** this_request, xacml_response_t ** this_response)
{
    size_t i, j, k, results_l, obligations_l, attrs_l;
    int curr_oblig;
    xacml_response_t * response = *this_response;
    xacml_result_t * result;
    xacml_obligation_t * obligation;
    xacml_attributeassignment_t * attr;
    xacml_fulfillon_t fulfillon;
    const char * obligation_id  = NULL;
    const char * attr_id        = NULL;
    size_t count_attrib_userid     = 0;
    size_t count_attrib_primgroupid= 0;
    size_t count_attrib_groupid    = 0;
    int primgroup_in_groups     = 0;
    const char * userid         = NULL;
    const char * groupid        = NULL;
    uid_t this_uid;
    gid_t this_prigid=(gid_t)(-1); /* initialize to silence compiler, although
				      only used when count_attrib_primgroupid>0
				    */
    gid_t *groupids=NULL;

    lcmaps_log(LOG_DEBUG, "%s\n", __func__);

    if (response == NULL) {
        lcmaps_log(LOG_DEBUG, "%s: response is NULL\n", __func__);
        return PEP_ERR_NULL_POINTER;
    }

    /* Check response message and verify that the amount of results and the
     * content of the results is OK */
    if (C_PEP_OK != checkResponseSanity (response)) {
	/* Don't log this on LOG_ERR: it can also be because of a Deny, errors
	 * are already logged by checkResponseSanity() itself too */
        lcmaps_log (LOG_INFO,
		"%s: checkResponseSanity() returned a failure condition in the response message. "
		"Stopped looking into the obligations\n", __func__);
        return MY_PEP_ERR_OH_PROCESS;
    } else {
        lcmaps_log(LOG_DEBUG,
		"%s: checkResponseSanity() returned OK. "
		"Continuing with the search for obligation information\n",
		__func__);
    }
        
    /* Going to obligations blob, skip the rest (as it was tested in detail in
     * checkResponseSanity()) */
    results_l = xacml_response_results_length(response);
    for (i = 0; i < results_l; i++) {
        result = xacml_response_getresult(response,(int)i);
        
        obligations_l = xacml_result_obligations_length(result);
        lcmaps_log(LOG_DEBUG, "%s: response.result[%lu]: %lu obligations\n",
		   __func__, (long unsigned)i, (long unsigned)obligations_l);
       
        /* Hunt for the obligation id of interest, skip the rest */
        for (j=0, curr_oblig=0; j < obligations_l; j++) {
	    /* Get current obligation: cannot use j, since we might remove
	     * obligations, resulting in a counter mismatch. */
            obligation = xacml_result_getobligation(result,curr_oblig);
	    obligation_id = xacml_obligation_getid(obligation);

            /*
             * Searching for the Obligation ID: 
             * XACML_GRIDWN_OBLIGATION_LOCAL_ENVIRONMENT_MAP_POSIX 
             */
            if (strcmp(XACML_GRIDWN_OBLIGATION_LOCAL_ENVIRONMENT_MAP_POSIX, obligation_id) != 0)
            {
                lcmaps_log(LOG_DEBUG,
			"%s: response.result[%lu].obligation[%lu].id = \"%s\". "
			"Skipping non-applicable Obligation ID\n",
                        __func__, (long unsigned)i, (long unsigned)j,
			obligation_id);
		curr_oblig++;
                continue;
            }

	    /* Check that we should run on Permit */
	    fulfillon=xacml_obligation_getfulfillon(obligation);
	    if (fulfillon!=XACML_FULFILLON_PERMIT)  {
		lcmaps_log(LOG_DEBUG,
			"%s:  response.result[%lu].obligation[%lu].id = \"%s\". "
			"Skipping obligation that should run on %s\n",
			__func__, (long unsigned)i, (long unsigned)j,
			obligation_id, fulfillon_str(fulfillon));
		curr_oblig++;
		continue;
	    }

            /* We've gottum! */
            lcmaps_log(LOG_DEBUG,
		    "%s: response.result[%lu].obligation[%lu].id = \"%s\". "
		    "Found applicable Obligation ID\n",
                    __func__, (long unsigned)i, (long unsigned)j, obligation_id);

            attrs_l = xacml_obligation_attributeassignments_length(obligation);
            lcmaps_log(LOG_DEBUG,
		    "%s: response.result[%lu].obligation[%lu]: %lu attributes.\n",
		    __func__, (long unsigned)i, (long unsigned)j,(long unsigned)attrs_l);

	    /* Allocate memory for group identifiers, maximally number of
	     * attributes */
	    if ( (groupids=calloc(attrs_l,sizeof(gid_t)))==NULL )   {
		lcmaps_log(LOG_ERR,"%s: out of memory\n",__func__);
		return MY_PEP_ERR_OH_PROCESS;
	    }
	   
	    /* Reset attribute counters per obligatio */
	    count_attrib_userid=count_attrib_primgroupid=count_attrib_groupid=0;

            for (k= 0; k < attrs_l; k++) {
                /* Reading individual attributes within obligations */
                attr = xacml_obligation_getattributeassignment(obligation, (int)k);
		attr_id = xacml_attributeassignment_getid(attr);

                /* Is the current value a User-ID */
                if (strcmp(XACML_GRIDWN_ATTRIBUTE_USER_ID, attr_id) == 0)
		{
                    /* Check if we didn't already had this attribute id... */
		    /* The maximum number of Usernames in an obligation is 1,
		     * and only 1 */
                    if (count_attrib_userid == 1) {
                        lcmaps_log (LOG_ERR,
				"%s: Error: more than 1 attribute of type \"%s\".\n", 
                                __func__, attr_id);
			goto failure;
                    }
                    
		    /* Got the username, now to convert it into UIDs through the
		     * POSIX account system */
		    userid = xacml_attributeassignment_getvalue(attr);
		    if (userid==NULL || *userid=='\0') {
			lcmaps_log (LOG_ERR,
			    "%s: Error: empty value at response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__,  (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
			errno = EINVAL;
			goto failure;
		    }
		   
		    /* Log id/value */
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].value= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, userid);

		    /* Converting the User-ID into a UID and registering the
		     * values into the LCMAPS framework */
		    if (threadsafe_getuid_from_name (userid, &this_uid) < 0) {
			lcmaps_log (LOG_ERR,
				"%s: Error: user-id not found: \"%s\".\n",
				__func__, userid);
			goto failure;
		    }

		    /* Add to internal list */
		    if (addUid(this_uid)<0)
			goto failure;
		    
		    /* Administrative, counting appearance of attribute-id
		     * XACML_GRIDWN_ATTRIBUTE_USER_ID */
                    count_attrib_userid++;
                }
                /* Is the current value a Primary Group-ID */
                else if (strcmp(XACML_GRIDWN_ATTRIBUTE_GROUP_ID_PRIMARY,attr_id) == 0)
		{
                    /* Check if we didn't already had this attribute id... */
		    /* The maximum number of Primary Groups in an obligation is
		     * 1, and only 1 */
                    if (count_attrib_primgroupid == 1) {
                        lcmaps_log(LOG_ERR,
				"%s: Error: more than 1 attribute of type \"%s\".\n", 
                                __func__, XACML_GRIDWN_ATTRIBUTE_GROUP_ID_PRIMARY);
			goto failure;
                    }
		    
		    /* Got the groupname, now to convert it into a GID through
		     * the POSIX account system */
		    groupid = xacml_attributeassignment_getvalue(attr);
		    if (groupid==NULL || *groupid=='\0') {
			lcmaps_log (LOG_ERR,
			    "%s: Error: empty value at response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__,  (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
			errno = EINVAL;
			goto failure;
		    }

		    /* Log id/value */
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].value= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, groupid);

		    /* Converting the Group-ID into a GID */
		    if (threadsafe_getgid_from_name(groupid, &this_prigid)<0)
		    {
			lcmaps_log (LOG_ERR,
				"%s: Error: group-id not found: \"%s\".\n",
				__func__, groupid);
			goto failure;
		    }
		    
		    /* Add to internal list */
		    if (addPGid(this_prigid)<0)
			goto failure;
		    
		    /* Administrative, counting appearance of attribute-id
		     * XACML_GRIDWN_ATTRIBUTE_GROUP_ID */
		    count_attrib_primgroupid++;
                }
                /* Is the current value a Group-ID (not necessarily secondary
		 * group! The profile specifies that the primary group should
		 * also appear in this list, we'll remove it later when adding
		 * the secondary GIDs */
                else if (strcmp(XACML_GRIDWN_ATTRIBUTE_GROUP_ID,attr_id) == 0)
		{
		    /* Got the groupname, now to convert it into a GID through
		     * the POSIX account system */
		    groupid = xacml_attributeassignment_getvalue(attr);
		    if (groupid==NULL || *groupid=='\0') {
			lcmaps_log (LOG_ERR,
			    "%s: Error: empty value at response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__,  (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
			errno = EINVAL;
			goto failure;
		    }

		    /* Log id/value */
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].id= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, attr_id);
		    lcmaps_log(LOG_DEBUG,
			    "%s: response.result[%lu].obligation[%lu].attribute[%lu].value= %s\n",
			    __func__, (long unsigned)i, (long unsigned)j, (long unsigned)k, groupid);

		    /* Converting the Group-ID into a GID */
		    if ( threadsafe_getgid_from_name(groupid, &(groupids[count_attrib_groupid]))<0)
		    {
			lcmaps_log (LOG_ERR,
				"%s: Error: group-id not found: \"%s\".\n",
				__func__, groupid);
			goto failure;
		    }
                    
		    /* Administrative, counting appearance of attribute-id
		     * XACML_GRIDWN_ATTRIBUTE_GROUP_ID_PRIMARY */
                    count_attrib_groupid++;
                }
		/* Unknown attribute */
                else
                {
                    lcmaps_log (LOG_ERR,
			    "%s: Error: Unexpected attribute: %s\n",
			    __func__, attr_id);
		    goto failure;
                }
            }

	    /* Check we found the user-id attribute, which is mandatory, see
	     * section 3.6.1, XACML Grid Worker Node Authorization Profile,
	     * Version 1.0 */
	    if (count_attrib_userid==0)	{
		lcmaps_log(LOG_ERR,
		    "%s: Error: obligation %s is missing required attribute\n",
		    __func__, obligation_id);
		goto failure;
            }
	
	    /* Sanity check on groups: primary group is optional, but if it
	     * appears, it must be in the list of groups, see section 3.6.2,
	     * XACML Grid Worker Node Authorization Profile, Version 1.0 */
	    if (count_attrib_primgroupid>0)	{
		/* Check whether primary GID appears in the list of groups */
		primgroup_in_groups=0;
		for (k=0; k<count_attrib_groupid; k++)	{
		    if (groupids[k]==this_prigid)   {
			primgroup_in_groups=1;
			break;
		    }
		}
		/* Check we found it */
		if (!primgroup_in_groups) {
		    lcmaps_log(LOG_ERR,
			    "%s: primary group %ld is not in the list of group-ids (%s)\n",
			    __func__, (long unsigned)this_prigid,
			    XACML_GRIDWN_ATTRIBUTE_GROUP_ID);
		    goto failure;
		}

		/* Add non-primary groups as secondary groups */
		for (k=0; k<count_attrib_groupid; k++)	{
		    /* if the sgid is not the pgid, add the group */
		    if (groupids[k]!=this_prigid) {
			/* Add to internal list */
			if (addSGid(&(groupids[k]),1)<0)
			    goto failure;
		    }
		}
	    } else {
		/* Add all groups as secondary groups */
		for (k=0; k<count_attrib_groupid; k++)	{
		    /* Add to internal list */
		    if (addSGid(&(groupids[k]),1)<0)
			goto failure;
		}
	    }

	    /* Free group list */
	    free(groupids); groupids=NULL;
		    
	    /* Now remove the handled obligation or update curr_oblig */
	    if (XacmlResultRemoveObligation)
		XacmlResultRemoveObligation(result,curr_oblig);
	    else
		curr_oblig++;
        }
    }

    /* All is fine */
    return PEP_OK;

failure:
    free(groupids); groupids=NULL;

    return MY_PEP_ERR_OH_PROCESS;
}

int oh_destroy_uidgid (void)
{
    lcmaps_log(LOG_DEBUG, "oh_destroy_uidgid()\n");
    return PEP_OK;
}

int oh_destroy_secondarygids (void)
{
    lcmaps_log(LOG_DEBUG, "oh_destroy_secondarygids()\n");
    return PEP_OK;
}

int oh_destroy_username(void)
{
    lcmaps_log(LOG_DEBUG, "oh_destroy_username()\n");
    return PEP_OK;
}

int oh_destroy_account(void)
{
    lcmaps_log(LOG_DEBUG, "oh_destroy_account()\n");
    return PEP_OK;
}

int oh_destroy_local_environment_map_posix (void)
{
    lcmaps_log(LOG_DEBUG, "oh_destroy_local_environment_map_posix()\n");
    return PEP_OK;
}
