/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

/* Next define is necessary for OpenSolaris for getpwnam_r() and getpwuid_r() */
#ifdef __sun
#define _POSIX_PTHREAD_SEMANTICS
#endif

#define _XOPEN_SOURCE	500

#include <lcmaps/lcmaps_account.h>
/* For lcmaps_get_gidlist */
#include <lcmaps/lcmaps_modules.h>

#include "pep-c-obligation-handlers_helpers.h"

#include <unistd.h>
#include <errno.h>

/* For gettimeofday() */
#include <sys/time.h>
/* For UINT_MAX */
#include <limits.h>


static list_supported_obligations_t * supp_obl_list = NULL;
static list_supported_obligations_t * supp_obl_list_index = NULL;

/* Current decision: need the information also for the round-robin strategy, can
 * be retrieved using pepc_get_decision(). Initialize to indeterminate */
static xacml_decision_t pepc_decision=XACML_DECISION_INDETERMINATE;

/* wether the pepc_decision has a valid value */
static int pepc_decision_valid=0;

/* variables to contain the internal list of credentials */
static uid_t *uid_list=NULL;
static gid_t *pgid_list=NULL;
static gid_t *sgid_list=NULL;
static size_t cnt_uid=0;
static size_t max_cnt_uid=0;
static size_t cnt_pgid=0;
static size_t max_cnt_pgid=0;
static size_t cnt_sgid=0;
static size_t max_cnt_sgid=0;

/* When set, a Not Applicable is interpreted as success. This is necessary when
 * Argus is only used for a banning list */
static int treat_notapplicable_as_success = 0;

/* setter function for the treat_notapplicable_as_success variable */
void set_treat_notapplicable_as_success(int val)
{
    treat_notapplicable_as_success = val;
}

/* getter function for the treat_notapplicable_as_success variable */
int get_treat_notapplicable_as_success(void)
{
    return treat_notapplicable_as_success;
}

/* reset function for the pepc_decision variable */
void pepc_reset_decision(void)
{
    pepc_decision_valid=0;
    pepc_decision=XACML_DECISION_INDETERMINATE;
}

/* getter function for the current pepc_decision variable */
int pepc_get_decision(xacml_decision_t *my_pepc_decision)
{
    if (pepc_decision_valid==0 || my_pepc_decision==NULL)
	return -1;

    *my_pepc_decision=pepc_decision;

    return 0;
}

char * get_next_supported_obligation_id (void)
{
    char * tmp = NULL;

    /* Should have an index, otherwise we're at the end of the list or no list
     * available */
    if (supp_obl_list_index)
    {
        tmp = supp_obl_list_index -> obligation_id;
        supp_obl_list_index = supp_obl_list_index -> next;

        return tmp;
    }
    else
        return NULL;
}


void rewind_supported_obligation_list (void)
{
    supp_obl_list_index = supp_obl_list;
}


int print_supported_obligations_list (void)
{
    list_supported_obligations_t * tmp = supp_obl_list;
    while (tmp)
    {
        lcmaps_log(LOG_DEBUG, "Registered supported obligations: %s\n", tmp -> obligation_id);
        tmp = tmp -> next;
    }
    return 0;
}


int add_supported_obligation_to_list (const char * obligation_id)
{
    list_supported_obligations_t * node = NULL;
    list_supported_obligations_t * tmp  = NULL;

    node = calloc ((size_t)1, sizeof(list_supported_obligations_t));

    if (!node)
    {
        return -1;
    }
    else
    {
        if (!(node -> obligation_id = strdup(obligation_id)))
        {
            free (node);
            return -1;
        }
        node -> next = NULL;

        if (!supp_obl_list)
        {
            /* Add node as first in the list */
            supp_obl_list_index = supp_obl_list = node;
        }
        else
        {
            tmp = supp_obl_list;

            /* Loop to final struct in the list */
            while (tmp -> next)
            { 
                tmp = tmp -> next; 
            }

            /* Add node to list */
            tmp -> next = node;
        }
    }

    /* print_supported_obligations_list(); */
    return 0;
}


int free_supported_obligations_list (void)
{
    list_supported_obligations_t * tmp = NULL;
    while (supp_obl_list)
    {
        tmp = supp_obl_list;
        supp_obl_list = supp_obl_list -> next;
        free(tmp->obligation_id);
        free(tmp);
    }
    return 0;
}


/**
 * Returns the string representation of the decision.
 */     
const char * decision_str(xacml_decision_t decision)
{
    switch (decision) 
    {
        case XACML_DECISION_DENY:
            return "Deny";
        case XACML_DECISION_PERMIT:
	    return "Permit";
        case XACML_DECISION_INDETERMINATE:
            return "Indeterminate";
        case XACML_DECISION_NOT_APPLICABLE:
            return "Not Applicable";
        default:
            return "ERROR: Unknown decision";
    }
}


/**
 * Returns the string representation of the fulfillon as obtained by
 * xacml_obligation_getfulfillon(obligation)
 */     
const char * fulfillon_str(xacml_fulfillon_t fulfillon)
{
    switch (fulfillon) 
    {
        case XACML_FULFILLON_DENY:
            return "Deny";
        case XACML_FULFILLON_PERMIT:
            return "Permit";
        default:
            return "ERROR: Unknown fulfillon";
    }
}


/*
 * This function will test if the response message contains the right 
 * amount of result sets and if the results are all OK and have a Permit 
 * decision 
 */
int checkResponseSanity (xacml_response_t * response)
{
    size_t i=0,results_l = 0;
    int failed = 0;
    const char *value = NULL;
    xacml_result_t * result = NULL;
    xacml_status_t * status = NULL;
    xacml_statuscode_t * statuscode = NULL;
    xacml_statuscode_t * subcode = NULL;


    lcmaps_log(LOG_DEBUG, "%s\n", __func__);

    if (response == NULL) {
        lcmaps_log(LOG_DEBUG, "%s: response is NULL\n", __func__);
        return PEP_ERR_NULL_POINTER;
    }

    results_l = xacml_response_results_length(response);
    lcmaps_log(LOG_DEBUG, "%s: response: %lu results\n",
	    __func__, (long unsigned)results_l);
    
    /* No results == fail! */
    if (results_l == 0) {
        lcmaps_log (LOG_ERR, "%s: Error: No results in response message. Communication failure between or unknown failure at the PEP Daemon\n", __func__);
        return C_PEP_FAIL;
    }

    /* Too many results == fail! */
    if (results_l > 1) {
        lcmaps_log (LOG_ERR, "%s: Error: %lu results in response message found. I can only handle one result from the PEP Daemon\n", __func__, (long unsigned)results_l);
        return C_PEP_FAIL;
    }


    /* Although I fail on anything else then one argument, for the sake of code
     * completeness I'll keep the for-loop on the amount of results */
    for (i = 0; i < results_l; i++) {
        result = xacml_response_getresult(response,(int)i);
	status = xacml_result_getstatus(result);
	statuscode = xacml_status_getcode(status);
	subcode = xacml_statuscode_getsubcode(statuscode);

	/* Note: decision is a static global and can also be retrieved using
	 * pepc_get_decision() */
	pepc_decision = xacml_result_getdecision(result);
	pepc_decision_valid = 1;

        /* Status code */
	value = xacml_statuscode_getvalue(statuscode);
        if (strcmp (XACML_STATUSCODE_OK,  value) == 0) {
            lcmaps_log(LOG_DEBUG,
		    "%s: response.result[%lu].status.code= %s\n",
		    __func__, (long unsigned)i, value);
	} else {
	    failed=1;
            lcmaps_log (LOG_WARNING,
		    "%s: response.result[%lu].status.code= %s\n",
                    __func__, (long unsigned)i, value);
        }

        /* Status sub code */
        if (subcode) {
	    value=xacml_statuscode_getvalue(subcode);
            if (strcmp (XACML_STATUSCODE_OK, value) == 0) {
                lcmaps_log(LOG_DEBUG,
			"%s: response.result[%lu].status.code.subcode= %s\n",
			__func__, (long unsigned)i, value);
            } else {
		failed=1;
                lcmaps_log (LOG_WARNING,
			"%s: response.result[%lu].status.code.subcode= %s\n",
			__func__, (long unsigned)i, value);
            }
        }

        /* Status message: log on warning in case status failed, otherwise on
	 * debug */
	value=xacml_status_getmessage(status);
	if (value)
	    lcmaps_log(failed ? LOG_WARNING : LOG_DEBUG,
		    "%s: response.result[%lu].status.message= %s\n",
		    __func__, (long unsigned)i, value);

	/* Decision */
	if (pepc_decision == XACML_DECISION_PERMIT ||
	    (treat_notapplicable_as_success &&
	     pepc_decision == XACML_DECISION_NOT_APPLICABLE
	    )
	   )
	{
	    lcmaps_log(LOG_DEBUG,
		    "%s: response.result[%lu].decision= %s\n",
		    __func__, (long unsigned)i, decision_str(pepc_decision));
	} else {
	    failed=1;
            lcmaps_log (LOG_ERR,
		    "%s: response.result[%lu].decision= %s\n",
                    __func__, (long unsigned)i, decision_str(pepc_decision));
        }

	/* If we failed return now */
	if (failed)
            return C_PEP_FAIL;

        /* Resource-id: unknown */
	value=xacml_result_getresourceid(result);
        lcmaps_log(LOG_DEBUG, "%s: response.result[%lu].resourceid= %s\n",
		__func__, (long unsigned)i, (value ? value : "NULL"));
        
    }

    /* All is ok */
    return C_PEP_OK;
}


int threadsafe_getuid_from_name (const char * username, uid_t *uid)
{
    struct passwd * p          = NULL;
    int             rc         = -1;
    int             errno_save = 0;
#ifndef HAVE_GETPWNAM_R /* Temporary hack */
    #define HAVE_GETPWNAM_R
#endif
#ifdef HAVE_GETPWNAM_R
    char *          pwtmpbuf = NULL;
    struct passwd   pwbuf;
    size_t          buflen = 1024;
  #ifdef _SC_GETPW_R_SIZE_MAX
    long	    sysconf_value = sysconf (_SC_GETPW_R_SIZE_MAX);;

    /* Check validity of parameters */
    if (username==NULL || uid==NULL)	{
	errno=EINVAL;
	return -1;
    }

    /* -1 can mean either invalid (errno!=0) or no limit (errno==0), in both
     * cases, use default buflen. In general, if the value is negative, we need
     * to use the default */
    if (sysconf_value < 0)
        buflen = 1024;
    else
	buflen = (size_t)sysconf_value;
  #endif

    /* Allocate the buffer for getpwnam_r() */
    pwtmpbuf = (char *) calloc ((size_t)1, buflen);

    /* Get Passwd Info for Username */
#ifdef __sun
    /* Somehow Solaris seems to use a non-POSIX version of getpwnam_r where the
     * buffersize is an int instead of a size_t */
    while (getpwnam_r (username, &pwbuf, pwtmpbuf, (int)buflen, &p) != 0)
#else
    while (getpwnam_r (username, &pwbuf, pwtmpbuf, buflen, &p) != 0)
#endif
    {
        /* If the errno is anything different then ERANGE, something is wrong or the ID is not available */
        if (errno != ERANGE)
        {
            errno_save = errno;
            p = NULL;
            break;
        }

        /* ERANGE is left over, which means that there is insufficient buffer space. Increasing buffer side and try again */
        buflen *= 2;
        free (pwtmpbuf);

        /* Allocate the buffer for getpwnam_r() */
        if (!(pwtmpbuf = (char *) calloc ((size_t)1, buflen)))
        {
            /* System ran out of memory */
            if (errno == ENOMEM)
            {
                /* No memory available, bail out */
                p = NULL;
                errno_save = errno;
                break;
            }
        }
    }
#else
    p = getpwnam (username);
#endif

    /* If there is an entry found, fetch the UID info */
    if (p)
    {
        *uid = p -> pw_uid;
	rc=0;
    }

    /* Remove the buffer for the getpwnam_r() */
    free (pwtmpbuf);
    errno = errno_save;
    return rc;
}


int threadsafe_getgid_from_name (const char * groupname, gid_t *gid)
{
    struct group  * g          = NULL;
    int             rc         = -1;
    int             errno_save = 0;
#ifndef HAVE_GETGRNAM_R /* Temporary hack */
    #define HAVE_GETGRNAM_R
#endif
#ifdef HAVE_GETGRNAM_R
    char *          grtmpbuf = NULL;
    struct group    grbuf;
    size_t          buflen = 1024;
  #ifdef _SC_GETGR_R_SIZE_MAX
    long	    sysconf_value = sysconf (_SC_GETGR_R_SIZE_MAX);;

    /* Check validity of parameters */
    if (groupname==NULL || gid==NULL)	{
	errno=EINVAL;
	return -1;
    }

    /* -1 can mean either invalid (errno!=0) or no limit (errno==0), in both
     * cases, use default buflen. In general, if the value is negative, we need
     * to use the default */
    if (sysconf_value < 0)
        buflen = 1024;
    else
	buflen = (size_t)sysconf_value;
  #endif

    /* Allocate the buffer for getpwnam_r() */
    grtmpbuf = (char *) calloc ((size_t)1, buflen);

    /* Get Passwd Info for Username */
#ifdef __sun
    /* Somehow Solaris seems to use a non-POSIX version of getgrnam_r where the
     * buffersize is an int instead of a size_t */
    while (getgrnam_r (groupname, &grbuf, grtmpbuf, (int)buflen, &g) != 0)
#else
    while (getgrnam_r (groupname, &grbuf, grtmpbuf, buflen, &g) != 0)
#endif
    {
        /* If the errno is anything different then ERANGE, something is wrong or the ID is not available */
        if (errno != ERANGE)
        {
            errno_save = errno;
            g = NULL;
            break;
        }

        /* ERANGE is left over, which means that there is insufficient buffer space. Increasing buffer side and try again */
        buflen *= 2;
        free (grtmpbuf);

        /* Allocate the buffer for getpwnam_r() */
        if (!(grtmpbuf = (char *) calloc ((size_t)1, buflen)))
        {
            /* System ran out of memory */
            if (errno == ENOMEM)
            {
                /* No memory available, bail out */
                g = NULL;
                errno_save = errno;
                break;
            }
        }
    }
#else
    /* If the platform doesn't have a getgrnam_r() function */
    g = getgrnam (username);
#endif

    /* If there is an entry found, fetch the GID info */
    if (g)
    {
        *gid = g -> gr_gid;
	rc=0;
    }

    /* Remove the buffer for the getgrnam_r() */
    free (grtmpbuf);
    errno = errno_save;
    return rc;
}



/*
 * Returns a random number via lrand48() cast to an unsigned.
 * The first time it will seed using the microseconds returned by gettimeofday
 * used in srand48(). Note that this function is not thread-safe.
 */
unsigned int pepc_rand(void)	{
    static int initialized=0;
    struct timeval tv;

    /* seed only once, use a fast changing variable, such as milliseconds */
    if (!initialized)	{
	gettimeofday(&tv,NULL);
	/* Create number of microseconds from tv_usec and tv_sec together, make
	 * sure to stay <= 2.147e9 < 2^31 */
	srand48((long)tv.tv_usec+1000000L*(long)(tv.tv_sec%2148));
	initialized=1;
    }

    /* return lrand48 modulo a max unsigned int */
    return (unsigned int)(lrand48() % (long)UINT_MAX);
}


/**
 * Returns list and number of UIDs as parameters,
 * returns -1 on error, 0 on success
 */
int get_uid_list(uid_t *uids[], size_t *count)	{
    if (uids==NULL || count==NULL)
	return -1;

    *uids=uid_list;
    *count=cnt_uid;

    return 0;
}

/**
 * Returns list and number of primary GIDs as parameters,
 * returns -1 on error, 0 on success
 */
int get_pgid_list(gid_t *pgids[], size_t *count)    {
    if (pgids==NULL || count==NULL)
	return -1;

    *pgids=pgid_list;
    *count=cnt_pgid;

    return 0;
}

/**
 * Returns list and number of secondary GIDs as parameters,
 * returns -1 on error, 0 on success
 */
int get_sgid_list(gid_t *sgids[], size_t *count)    {
    if (sgids==NULL || count==NULL)
	return -1;

    *sgids=sgid_list;
    *count=cnt_sgid;

    return 0;
}

/**
 * Free and re-initialize internal lists of UIDs, pGIDs and sGIDs
 */
void clean_uidgid_lists(void)	{
    free(uid_list); uid_list=NULL;
    free(pgid_list); pgid_list=NULL;
    free(sgid_list); sgid_list=NULL;

    cnt_uid=max_cnt_uid=0;
    cnt_pgid=max_cnt_pgid=0;
    cnt_sgid=max_cnt_sgid=0;
}


/**
 * Returns list and number of UIDs as parameters,
 * returns -1 on error, 0 on success
 */
int addUid(uid_t uid)    {
    uid_t *new_uid_list;
    size_t new_max_cnt_uid=max_cnt_uid;

    /* Increase global list when needed */
    if (cnt_uid>=max_cnt_uid)	{
	new_max_cnt_uid=max_cnt_uid+10;
	new_uid_list=(uid_t*)realloc(uid_list, new_max_cnt_uid*sizeof(uid_t));
	if (new_uid_list==NULL)	{
	    lcmaps_log(LOG_ERR,
		    "%s: out of memory trying to add uid %ld\n",
		    __func__, (long)uid);
	    return -1;
	}
	max_cnt_uid=new_max_cnt_uid;
	uid_list=new_uid_list;
    }
   
    /* Add uid */
    lcmaps_log(LOG_DEBUG, "%s: Adding UID:            %ld\n",
	       __func__, (long)uid);
    uid_list[cnt_uid++]=uid;

    return 0;
}

/**
 * Adds given gid to internal list of received primary GIDs
 */
int addPGid(gid_t pgid)	{
    gid_t *new_pgid_list;
    size_t new_max_cnt_pgid=max_cnt_pgid;

    /* Increase global list when needed */
    if (cnt_pgid>=max_cnt_pgid)	{
	new_max_cnt_pgid=max_cnt_pgid+10;
	new_pgid_list=(gid_t*)realloc(pgid_list, new_max_cnt_pgid*sizeof(gid_t));
	if (new_pgid_list==NULL)	{
	    lcmaps_log(LOG_ERR,
		    "%s: out of memory trying to add pgid %ld\n",
		    __func__,(long)pgid);
	    return -1;
	}
	max_cnt_pgid=new_max_cnt_pgid;
	pgid_list=new_pgid_list;
    }
   
    /* Add pgid */
    lcmaps_log(LOG_DEBUG, "%s: Adding primary GID:    %ld\n",
	       __func__, (long)pgid);
    pgid_list[cnt_pgid++]=pgid;

    return 0;
}

/**
 * Adds given list of gids to internal list of received secondary GIDs
 */
int addSGid(gid_t sgid[], size_t count)	{
    gid_t *new_sgid_list;
    size_t i, new_max_cnt_sgid=max_cnt_sgid;

    /* Increase global list when needed */
    if (cnt_sgid+count>max_cnt_sgid)	{
	new_max_cnt_sgid=max_cnt_sgid+count+10;
	new_sgid_list=(gid_t*)realloc(sgid_list, new_max_cnt_sgid*sizeof(gid_t));
	if (new_sgid_list==NULL)	{
	    lcmaps_log(LOG_ERR,
		    "%s: out of memory trying to add %ld sgid(s)\n",
		    __func__, (long)count);
	    return -1;
	}
	max_cnt_sgid=new_max_cnt_sgid;
	sgid_list=new_sgid_list;
    }
   
    /* Add sgid */
    for (i=0; i<count; i++) {
        lcmaps_log(LOG_DEBUG, "%s: Adding secondary GID:  %ld\n",
		   __func__, (long)sgid[i]);
	sgid_list[cnt_sgid++]=sgid[i];
    }

    return 0;
}
